# Copyright (c) 2018-2024 Tigera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import subprocess
import json
import sys
import time

from tests.k8st.test_base import TestBase
from tests.k8st.utils.utils import start_external_node_with_bgp, \
        retry_until_success, run, curl, DiagsCollector, calicoctl, kubectl, node_info, NGINX_IMAGE

_log = logging.getLogger(__name__)

attempts = 10

bird_conf = """
# Template for all BGP clients
template bgp bgp_template {
  debug { states };
  description "Connection to BGP peer";
  local as 64512;
  multihop;
  gateway recursive; # This should be the default, but just in case.
  import all;        # Import all routes, since we don't know what the upstream
                     # topology is and therefore have to trust the ToR/RR.
  export all;
  source address ip@local;  # The local address we use for the TCP connection
  add paths on;
  graceful restart;  # See comment in kernel section about graceful restart.
  connect delay time 2;
  connect retry time 5;
  error wait time 5,30;
}

# ------------- Node-to-node mesh -------------
protocol bgp Mesh_with_master_node from bgp_template {
  neighbor %s as 64512;
  passive on; # Mesh is unidirectional, peer will connect to us.
  password "very-secret";
}

protocol bgp Mesh_with_node_1 from bgp_template {
  neighbor %s as 64512;
  passive on; # Mesh is unidirectional, peer will connect to us.
  password "very-secret";
}

protocol bgp Mesh_with_node_2 from bgp_template {
  neighbor %s as 64512;
  passive on; # Mesh is unidirectional, peer will connect to us.
  password "very-secret";
}

protocol bgp Mesh_with_node_3 from bgp_template {
  neighbor %s as 64512;
  passive on; # Mesh is unidirectional, peer will connect to us.
  password "very-secret";
}
"""

# BIRD config for an external node to peer with
# the in-cluster route reflector on kube-node-2.
bird_conf_rr = """
# Template for all BGP clients
template bgp bgp_template {
  debug { states };
  description "Connection to BGP peer";
  local as 64512;
  multihop;
  gateway recursive; # This should be the default, but just in case.
  import all;        # Import all routes, since we don't know what the upstream
                     # topology is and therefore have to trust the ToR/RR.
  export all;
  source address ip@local;  # The local address we use for the TCP connection
  add paths on;
  graceful restart;  # See comment in kernel section about graceful restart.
  connect delay time 2;
  connect retry time 5;
  error wait time 5,30;
}

protocol bgp Mesh_with_node_2 from bgp_template {
  neighbor %s as 64512;
  passive on; # Mesh is unidirectional, peer will connect to us.
}
"""

class _TestBGPAdvert(TestBase):

    def setUp(self):
        super(_TestBGPAdvert, self).setUp()

        # Create bgp test namespace
        self.ns = "bgp-test"
        self.create_namespace(self.ns)

        self.nodes, self.ips, _ = node_info()
        self.external_node_ip = start_external_node_with_bgp(
            "kube-node-extra",
            bird_peer_config=self.get_bird_conf(),
        )

        # Establish BGPPeer from cluster nodes to node-extra
        calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPPeer
metadata:
  name: node-extra.peer%s
EOF
""" % self.get_extra_peer_spec())

        kubectl("""apply -f - <<EOF
apiVersion: v1
kind: Secret
metadata:
  name: bgp-secrets
  namespace: calico-system
type: Opaque
stringData:
  rr-password: very-secret
EOF
""")

    def tearDown(self):
        super(_TestBGPAdvert, self).tearDown()
        self.delete_and_confirm(self.ns, "ns")
        try:
            # Delete the extra node.
            run("docker rm -f kube-node-extra")
        except subprocess.CalledProcessError:
            pass

        # Delete BGPPeers.
        calicoctl("delete bgppeer node-extra.peer", allow_fail=True)
        calicoctl("delete bgppeer peer-with-rr", allow_fail=True)

        # Restore node-to-node mesh.
        calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata: {name: default}
spec:
  nodeToNodeMeshEnabled: true
  asNumber: 64512
EOF
""")

        # Remove node-2's route-reflector config, using a retry to avoid
        # transient conflict errors.
        retry_until_success(lambda: self.clear_rr_config(self.nodes[2]))

    def set_rr_config(self, node):
        json_str = calicoctl("get node %s -o json" % node)
        node_dict = json.loads(json_str)
        node_dict['metadata']['labels']['i-am-a-route-reflector'] = 'true'
        node_dict['spec']['bgp']['routeReflectorClusterID'] = '224.0.0.1'
        calicoctl("""apply -f - << EOF
%s
EOF
""" % json.dumps(node_dict))

    def clear_rr_config(self, node):
        json_str = calicoctl("get node %s -o json" % node)
        node_dict = json.loads(json_str)
        node_dict['metadata']['labels'].pop('i-am-a-route-reflector', '')
        node_dict['spec']['bgp'].pop('routeReflectorClusterID', '')
        calicoctl("""apply -f - << EOF
%s
EOF
""" % json.dumps(node_dict))

    def get_svc_cluster_ip(self, svc, ns):
        return kubectl("get svc %s -n %s -o json | jq -r .spec.clusterIP" %
                       (svc, ns)).strip()

    def get_svc_loadbalancer_ip(self, svc, ns):
        for i in range(10):
            lb_ip = kubectl("get svc %s -n %s -o json | jq -r .status.loadBalancer.ingress[0].ip" %
                       (svc, ns)).strip()
            if lb_ip != "null":
                return lb_ip
            time.sleep(1)
        raise Exception("No LoadBalancer IP found for service: %s/%s" % (ns, svc))

    def assert_ecmp_routes(self, dst, via):
        matchStr = dst + " proto bird "
        # sort ips and construct match string for ECMP routes.
        for ip in sorted(via):
            matchStr += "\n\tnexthop via %s dev eth0 weight 1 " % ip
        retry_until_success(lambda: self.assertIn(matchStr, self.get_routes()))

    def get_svc_host_ip(self, svc, ns):
        return kubectl("get po -l app=%s -n %s -o json | jq -r .items[0].status.hostIP" %
                       (svc, ns)).strip()

    def add_svc_external_ips(self, svc, ns, ips):
        ipsStr = ','.join('"{0}"'.format(ip) for ip in ips)
        patchStr = "{\"spec\": {\"externalIPs\": [%s]}}" % (ipsStr)
        return kubectl("patch svc %s -n %s --patch '%s'" % (svc, ns, patchStr)).strip()


class TestBGPAdvert(_TestBGPAdvert):

    # In the tests of this class we have a full BGP mesh between the
    # cluster nodes (kube-control-plane, kube-node-1 and kube-node-2)
    # and the external node (kube-node-extra):
    #
    # - The full mesh between the cluster nodes is configured by
    #   nodeToNodeMeshEnabled: true.
    #
    # - The peerings from each cluster node to the external node are
    #   configured by self.get_extra_peer_spec().
    #
    # - The peerings from the external node to each cluster node are
    #   configured by self.get_bird_conf().

    def get_bird_conf(self):
        return bird_conf % (self.ips[0], self.ips[1], self.ips[2], self.ips[3])

    def get_extra_peer_spec(self):
        return """
spec:
  peerIP: %s
  asNumber: 64512
  password:
    secretKeyRef:
      name: bgp-secrets
      key: rr-password
""" % self.external_node_ip

    def test_cluster_ip_advertisement(self):
        """
        Runs the tests for service cluster IP advertisement
        - Create both a Local and a Cluster type NodePort service with a single replica.
          - assert only local and cluster CIDR routes are advertised.
          - assert /32 routes are used, source IP is preserved.
        - Scale the Local NP service so it is running on multiple nodes, assert ECMP routing, source IP is preserved.
        - Delete both services, assert only cluster CIDR route is advertised.
        """
        with DiagsCollector():

            calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata:
  name: default
spec:
  serviceClusterIPs:
  - cidr: 10.96.0.0/12
EOF
""")

            # Assert that a route to the service IP range is present.
            retry_until_success(lambda: self.assertIn("10.96.0.0/12", self.get_routes()))

            # Create both a Local and a Cluster type NodePort service with a single replica.
            local_svc = "nginx-local"
            cluster_svc = "nginx-cluster"
            self.deploy(NGINX_IMAGE, local_svc, self.ns, 80)
            self.deploy(NGINX_IMAGE, cluster_svc, self.ns, 80, traffic_policy="Cluster")
            self.wait_until_exists(local_svc, "svc", self.ns)
            self.wait_until_exists(cluster_svc, "svc", self.ns)

            # Get clusterIPs.
            local_svc_ip = self.get_svc_cluster_ip(local_svc, self.ns)
            cluster_svc_ip = self.get_svc_cluster_ip(cluster_svc, self.ns)

            # Wait for the deployments to roll out.
            self.wait_for_deployment(local_svc, self.ns)
            self.wait_for_deployment(cluster_svc, self.ns)

            # Assert that both nginx service can be curled from the external node.
            retry_until_success(curl, function_args=[local_svc_ip])
            retry_until_success(curl, function_args=[cluster_svc_ip])

            # Assert that local clusterIP is an advertised route and cluster clusterIP is not.
            retry_until_success(lambda: self.assertIn(local_svc_ip, self.get_routes()))
            retry_until_success(lambda: self.assertNotIn(cluster_svc_ip, self.get_routes()))


            # Scale the local_svc to 4 replicas
            self.scale_deployment(local_svc, self.ns, 4)
            self.wait_for_deployment(local_svc, self.ns)
            self.assert_ecmp_routes(local_svc_ip, [self.ips[1], self.ips[2], self.ips[3]])
            for i in range(attempts):
              retry_until_success(curl, function_args=[local_svc_ip])

            # Delete both services.
            self.delete_and_confirm(local_svc, "svc", self.ns)
            self.delete_and_confirm(cluster_svc, "svc", self.ns)

            # Assert that clusterIP is no longer an advertised route.
            retry_until_success(lambda: self.assertNotIn(local_svc_ip, self.get_routes()))

    def test_node_exclusion(self):
        """
        Tests the node exclusion label.
        - Create services, assert advertised from all nodes.
        - Label one node so that it is excluded, assert that routes are withdrawn from that node.
        - Delete / recreate service, assert it is still advertised from the correct nodes.
        - Remove the exclusion label, assert that the node re-advertises the svc.
        """
        with DiagsCollector():

            calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata:
  name: default
spec:
  serviceClusterIPs:
  - cidr: 10.96.0.0/12
  serviceExternalIPs:
  - cidr: 175.200.0.0/16
EOF
""")

            # Assert that a route to the service IP range is present.
            cluster_cidr = "10.96.0.0/12"
            retry_until_success(lambda: self.assertIn(cluster_cidr, self.get_routes()))

            # Create both a Local and a Cluster type NodePort service with a single replica.
            local_svc = "nginx-local"
            cluster_svc = "nginx-cluster"
            self.deploy(NGINX_IMAGE, local_svc, self.ns, 80)
            self.deploy(NGINX_IMAGE, cluster_svc, self.ns, 80, traffic_policy="Cluster")
            self.wait_until_exists(local_svc, "svc", self.ns)
            self.wait_until_exists(cluster_svc, "svc", self.ns)

            # Get clusterIPs.
            local_svc_ip = self.get_svc_cluster_ip(local_svc, self.ns)
            cluster_svc_ip = self.get_svc_cluster_ip(cluster_svc, self.ns)

            # Wait for the deployments to roll out.
            self.wait_for_deployment(local_svc, self.ns)
            self.wait_for_deployment(cluster_svc, self.ns)

            # Assert that both nginx service can be curled from the external node.
            retry_until_success(curl, function_args=[local_svc_ip])
            retry_until_success(curl, function_args=[cluster_svc_ip])

            # Assert that local clusterIP is an advertised route and cluster clusterIP is not.
            retry_until_success(lambda: self.assertIn(local_svc_ip, self.get_routes()))
            retry_until_success(lambda: self.assertNotIn(cluster_svc_ip, self.get_routes()))

            # Connectivity should always succeed.
            for i in range(attempts):
              retry_until_success(curl, function_args=[local_svc_ip])
              retry_until_success(curl, function_args=[cluster_svc_ip])

            # Scale local service to 4 replicas
            self.scale_deployment(local_svc, self.ns, 4)
            self.wait_for_deployment(local_svc, self.ns)
            self.wait_for_deployment(cluster_svc, self.ns)

            # Assert routes are correct and services are accessible.
            # Local service should only be advertised from nodes that can run pods.
            # The cluster CIDR should be advertised from all nodes.
            self.assert_ecmp_routes(local_svc_ip, [self.ips[1], self.ips[2], self.ips[3]])
            self.assert_ecmp_routes(cluster_cidr, [self.ips[0], self.ips[1], self.ips[2], self.ips[3]])
            for i in range(attempts):
              retry_until_success(curl, function_args=[local_svc_ip])

            # Label one node in order to exclude it from service advertisement.
            # After this, we should expect that all routes from that node are
            # withdrawn.
            kubectl("label node %s node.kubernetes.io/exclude-from-external-load-balancers=true --overwrite" % self.nodes[1])

            # Assert routes are correct and services are accessible.
            # It should no longer have a route via self.nodes[1]
            self.assert_ecmp_routes(local_svc_ip, [self.ips[2], self.ips[3]])
            self.assert_ecmp_routes(cluster_cidr, [self.ips[0], self.ips[2], self.ips[3]])

            # Should work the same for external IP cidr.
            external_ip_cidr = "175.200.0.0/16"
            self.assert_ecmp_routes(external_ip_cidr, [self.ips[0], self.ips[2], self.ips[3]])

            # Should still be reachable through other nodes.
            for i in range(attempts):
              retry_until_success(curl, function_args=[local_svc_ip])
              retry_until_success(curl, function_args=[cluster_svc_ip])

            # Delete the local service, confirm that it is no longer advertised.
            self.delete_and_confirm(local_svc, "svc", self.ns)
            retry_until_success(lambda: self.assertNotIn(local_svc_ip, self.get_routes()))

            # Re-create the local service. Assert it is advertised from the correct nodes,
            # but not from the excluded node.
            self.create_service(local_svc, local_svc, self.ns, 80)
            self.wait_until_exists(local_svc, "svc", self.ns)
            local_svc_ip = self.get_svc_cluster_ip(local_svc, self.ns)
            self.assert_ecmp_routes(local_svc_ip, [self.ips[2], self.ips[3]])
            for i in range(attempts):
              retry_until_success(curl, function_args=[local_svc_ip])

            # Add an external IP to the local svc and assert it follows the same
            # advertisement rules.
            local_svc_external_ip = "175.200.1.1"
            self.add_svc_external_ips(local_svc, self.ns, [local_svc_external_ip])
            self.assert_ecmp_routes(local_svc_external_ip, [self.ips[2], self.ips[3]])

            # Enable the excluded node. Assert that the node starts
            # advertising service routes again.
            kubectl("label node %s node.kubernetes.io/exclude-from-external-load-balancers=false --overwrite" % self.nodes[1])
            self.assert_ecmp_routes(local_svc_ip, [self.ips[1], self.ips[2], self.ips[3]])
            self.assert_ecmp_routes(local_svc_external_ip, [self.ips[1], self.ips[2], self.ips[3]])
            self.assert_ecmp_routes(cluster_cidr, [self.ips[0], self.ips[1], self.ips[2], self.ips[3]])
            for i in range(attempts):
              retry_until_success(curl, function_args=[local_svc_ip])

            # Delete both services.
            self.delete_and_confirm(local_svc, "svc", self.ns)
            self.delete_and_confirm(cluster_svc, "svc", self.ns)

            # Assert that clusterIP is no longer an advertised route.
            retry_until_success(lambda: self.assertNotIn(local_svc_ip, self.get_routes()))

    def test_external_ip_advertisement(self):
        """
        Runs the tests for service external IP advertisement
        """
        with DiagsCollector():

            # Allow two IP ranges for the external IPs we'll test with
            calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata:
  name: default
spec:
  serviceExternalIPs:
  - cidr: 175.200.0.0/16
  - cidr: 200.255.0.0/24
EOF
""")

            # Create both a Local and a Cluster type NodePort service with a single replica.
            local_svc = "nginx-local"
            cluster_svc = "nginx-cluster"
            self.deploy(NGINX_IMAGE, local_svc, self.ns, 80)
            self.deploy(NGINX_IMAGE, cluster_svc, self.ns, 80, traffic_policy="Cluster")
            self.wait_until_exists(local_svc, "svc", self.ns)
            self.wait_until_exists(cluster_svc, "svc", self.ns)

            # Get clusterIPs.
            local_svc_ip = self.get_svc_cluster_ip(local_svc, self.ns)
            cluster_svc_ip = self.get_svc_cluster_ip(cluster_svc, self.ns)

            # Wait for the deployments to roll out.
            self.wait_for_deployment(local_svc, self.ns)
            self.wait_for_deployment(cluster_svc, self.ns)

            # Assert that clusterIPs are not advertised.
            retry_until_success(lambda: self.assertNotIn(local_svc_ip, self.get_routes()))
            retry_until_success(lambda: self.assertNotIn(cluster_svc_ip, self.get_routes()))

            # Create a network policy that only accepts traffic from the external node.
            kubectl("""apply -f - << EOF
apiVersion: networking.k8s.io/v1
kind: NetworkPolicy
metadata:
  name: allow-tcp-80-ex
  namespace: bgp-test
spec:
  podSelector: {}
  policyTypes:
  - Ingress
  ingress:
  - from:
    - ipBlock: { cidr: %s/32 }
    ports:
    - protocol: TCP
      port: 80
EOF
""" % self.external_node_ip)

            # Get host IPs for the nginx pods.
            local_svc_host_ip = self.get_svc_host_ip(local_svc, self.ns)
            cluster_svc_host_ip = self.get_svc_host_ip(cluster_svc, self.ns)

            # Select an IP from each external IP CIDR.
            local_svc_external_ip = "175.200.1.1"
            cluster_svc_external_ip = "200.255.255.1"

            # Add external IPs to the two services.
            self.add_svc_external_ips(local_svc, self.ns, [local_svc_external_ip])
            self.add_svc_external_ips(cluster_svc, self.ns, [cluster_svc_external_ip])

            # Verify that external IPs for local service is advertised but not the cluster service.
            local_svc_externalips_route = "%s via %s" % (local_svc_external_ip, local_svc_host_ip)
            cluster_svc_externalips_route = "%s via %s" % (cluster_svc_external_ip, cluster_svc_host_ip)
            retry_until_success(lambda: self.assertIn(local_svc_externalips_route, self.get_routes()))
            retry_until_success(lambda: self.assertNotIn(cluster_svc_externalips_route, self.get_routes()))

            # Scale the local_svc to 4 replicas.
            self.scale_deployment(local_svc, self.ns, 4)
            self.wait_for_deployment(local_svc, self.ns)

            # Verify that we have ECMP routes for the external IP of the local service.
            self.assert_ecmp_routes(local_svc_external_ip, [self.ips[1], self.ips[2], self.ips[3]])

            # Delete both services, assert only cluster CIDR route is advertised.
            self.delete_and_confirm(local_svc, "svc", self.ns)
            self.delete_and_confirm(cluster_svc, "svc", self.ns)

            # Assert that external IP is no longer an advertised route.
            retry_until_success(lambda: self.assertNotIn(local_svc_externalips_route, self.get_routes()))

    def test_fully_qualified_service_ips(self):
        """
        Test advertisement of /32 Service IPs.
        """
        with DiagsCollector():

            # Allow exact IPs for each type of service IP. We expect these to be properly advertised.
            calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata:
  name: default
spec:
  serviceExternalIPs:
  - cidr: 90.15.0.1/32
EOF
""")

            # Create a Service with the External IP above, using
            # externalTrafficPolicy=Cluster. This should trigger advertisement
            # from all nodes.
            svc_name = "nginx-svc"
            ext_ip = "90.15.0.1"
            self.deploy(NGINX_IMAGE, svc_name, self.ns, 80, traffic_policy="Cluster", svc_type="ClusterIP", ext_ip=ext_ip)
            self.wait_until_exists(svc_name, "svc", self.ns)
            self.wait_for_deployment(svc_name, self.ns)

            # Verify the ext IP address is advertised from all nodes.
            self.assert_ecmp_routes(ext_ip, [self.ips[0], self.ips[1], self.ips[2], self.ips[3]])

    def test_loadbalancer_ip_advertisement(self):
        """
        Runs the tests for service LoadBalancer IP advertisement
        """
        with DiagsCollector():

            # Allow IP ranges for the LB IPs we'll test with
            calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata:
  name: default
spec:
  serviceLoadBalancerIPs:
  - cidr: 80.15.0.0/24
EOF
""")

            # Create a dummy service first to occupy the first LB IP. This is
            # a hack to make sure the chosen IP we use in the tests below
            # isn't the same as the zero address in the range.
            self.create_service("dummy-service", "dummy-service", self.ns, 80, svc_type="LoadBalancer")

            # Create both a Local and a Cluster type NodePort service with a single replica.
            local_svc = "nginx-local"
            cluster_svc = "nginx-cluster"
            self.deploy(NGINX_IMAGE, cluster_svc, self.ns, 80, traffic_policy="Cluster", svc_type="LoadBalancer")
            self.deploy(NGINX_IMAGE, local_svc, self.ns, 80, svc_type="LoadBalancer")
            self.wait_until_exists(local_svc, "svc", self.ns)
            self.wait_until_exists(cluster_svc, "svc", self.ns)

            # Get the allocated LB IPs.
            local_lb_ip = self.get_svc_loadbalancer_ip(local_svc, self.ns)
            cluster_lb_ip = self.get_svc_loadbalancer_ip(cluster_svc, self.ns)

            # Wait for the deployments to roll out.
            self.wait_for_deployment(local_svc, self.ns)
            self.wait_for_deployment(cluster_svc, self.ns)

            # Get host IPs for the nginx pods.
            local_svc_host_ip = self.get_svc_host_ip(local_svc, self.ns)
            cluster_svc_host_ip = self.get_svc_host_ip(cluster_svc, self.ns)

            # Verify that LB IP for local service is advertised but not the cluster service.
            local_svc_lb_route = "%s via %s" % (local_lb_ip, local_svc_host_ip)
            cluster_svc_lb_route = "%s via %s" % (cluster_lb_ip, cluster_svc_host_ip)
            retry_until_success(lambda: self.assertIn(local_svc_lb_route, self.get_routes()))
            retry_until_success(lambda: self.assertNotIn(cluster_svc_lb_route, self.get_routes()))

            # The full range should be advertised from each node.
            lb_cidr = "80.15.0.0/24"
            self.assert_ecmp_routes(lb_cidr, [self.ips[0], self.ips[1], self.ips[2], self.ips[3]])

            # Scale the local_svc to 4 replicas.
            self.scale_deployment(local_svc, self.ns, 4)
            self.wait_for_deployment(local_svc, self.ns)

            # Verify that we have ECMP routes for the LB IP of the local service from nodes running it.
            self.assert_ecmp_routes(local_lb_ip, [self.ips[1], self.ips[2], self.ips[3]])

            # Apply a modified BGP config that no longer enables advertisement
            # for LoadBalancer IPs.
            calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata:
  name: default
spec: {}
EOF
""")
            # Assert routes are withdrawn.
            retry_until_success(lambda: self.assertNotIn(local_lb_ip, self.get_routes()))
            retry_until_success(lambda: self.assertNotIn(lb_cidr, self.get_routes()))

            # Apply a modified BGP config that has a mismatched CIDR specified.
            calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata:
  name: default
spec:
  serviceLoadBalancerIPs:
  - cidr: 90.15.0.0/24
EOF
""")
            # Assert routes are still withdrawn.
            retry_until_success(lambda: self.assertNotIn(local_lb_ip, self.get_routes()))
            retry_until_success(lambda: self.assertNotIn(lb_cidr, self.get_routes()))

            # Reapply the correct configuration, we should see routes come back.
            calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata:
  name: default
spec:
  serviceLoadBalancerIPs:
  - cidr: 80.15.0.0/24
EOF
""")
            # Verify that we have ECMP routes for the LB IP of the local service from nodes running it.
            self.assert_ecmp_routes(local_lb_ip, [self.ips[1], self.ips[2], self.ips[3]])
            retry_until_success(lambda: self.assertIn(lb_cidr, self.get_routes()))
            retry_until_success(lambda: self.assertNotIn(cluster_svc_lb_route, self.get_routes()))

            # Services should be reachable from the external node.
            retry_until_success(curl, function_args=[local_lb_ip])
            retry_until_success(curl, function_args=[cluster_lb_ip])

            # Delete both services, assert only CIDR route is advertised.
            self.delete_and_confirm(local_svc, "svc", self.ns)
            self.delete_and_confirm(cluster_svc, "svc", self.ns)

            # Assert that LB IP is no longer an advertised route.
            retry_until_success(lambda: self.assertNotIn(local_lb_ip, self.get_routes()))

    def test_many_services(self):
        """
        Creates a lot of services quickly
        """
        with DiagsCollector():

            calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata:
  name: default
spec:
  serviceClusterIPs:
  - cidr: 10.96.0.0/12
EOF
""")

            # Assert that a route to the service IP range is present.
            retry_until_success(lambda: self.assertIn("10.96.0.0/12", self.get_routes()))

            # Create a local service and deployment.
            local_svc = "nginx-local"
            self.deploy(NGINX_IMAGE, local_svc, self.ns, 80)
            self.wait_for_deployment(local_svc, self.ns)

            # Get clusterIPs.
            cluster_ips = []
            cluster_ips.append(self.get_svc_cluster_ip(local_svc, self.ns))

            # Create many more services which select this deployment.
            num_svc = 300
            for i in range(num_svc):
                name = "nginx-svc-%s" % i
                self.create_service(name, local_svc, self.ns, 80)

            # Get all of their IPs.
            for i in range(num_svc):
                name = "nginx-svc-%s" % i
                cluster_ips.append(self.get_svc_cluster_ip(name, self.ns))

            # Assert they are all advertised to the other node. This should happen
            # quickly enough that by the time we have queried all services from
            # the k8s API, they should be programmed on the remote node.
            def check_routes_advertised():
                routes = self.get_routes()
                for cip in cluster_ips:
                    self.assertIn(cip, routes)
            retry_until_success(check_routes_advertised, retries=3, wait_time=5)

            # Scale to 0 replicas, assert all routes are removed.
            self.scale_deployment(local_svc, self.ns, 0)
            self.wait_for_deployment(local_svc, self.ns)
            def check_routes_gone():
                routes = self.get_routes()
                for cip in cluster_ips:
                    self.assertNotIn(cip, routes)
            retry_until_success(check_routes_gone, retries=10, wait_time=5)

    def test_bgp_filter_ip_advertisement(self):
        with DiagsCollector():
            # Add BGPConfiguration with serviceClusterIP
            kubectl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata:
  name: default
spec:
  serviceClusterIPs:
  - cidr: 10.96.0.0/12
EOF
""")
            # Assert that a route to the service IP range is present.
            retry_until_success(lambda: self.assertIn("10.96.0.0/12", self.get_routes()))

            # Create a Local type NodePort service with a single replica.
            local_svc = "nginx-local"
            self.deploy(NGINX_IMAGE, local_svc, self.ns, 80)
            self.wait_until_exists(local_svc, "svc", self.ns)

            # Get clusterIPs.
            local_svc_ip = self.get_svc_cluster_ip(local_svc, self.ns)

            # Wait for the deployment to roll out.
            self.wait_for_deployment(local_svc, self.ns)

            # Assert that nginx service can be curled from the external node.
            retry_until_success(curl, function_args=[local_svc_ip])

            # Assert that local clusterIP is an advertised route.
            retry_until_success(lambda: self.assertIn(local_svc_ip, self.get_routes()))

            # Create an export BGP filter that rejects the service IP range
            kubectl("""apply -f - <<EOF
apiVersion: projectcalico.org/v3
kind: BGPFilter
metadata:
  name: test-filter-export-1
spec:
  exportV4:
  - cidr: 10.96.0.0/12
    matchOperator: In
    action: Reject
EOF
""")
        kubectl("patch --type=merge bgppeer node-extra.peer --patch '{\"spec\": {\"filters\": [\"test-filter-export-1\"]}}'")
        self.add_cleanup(lambda: kubectl("patch --type=merge bgppeer node-extra.peer --patch '{\"spec\": {\"filters\": []}}'"))
        self.add_cleanup(lambda: kubectl("delete bgpfilter test-filter-export-1"))

        # Assert that local clusterIP is no longer advertised.
        retry_until_success(lambda: self.assertNotIn(local_svc_ip, self.get_routes()))
        # Assert that a route to the service IP range is no longer present.
        retry_until_success(lambda: self.assertNotIn("10.96.0.0/12", self.get_routes()))

class TestBGPAdvertRR(_TestBGPAdvert):

    # In the tests of this class, kube-node-2 acts as an RR, and all
    # the other nodes peer with it.  Here are the peerings that we
    # need for that:
    #
    #                                      RR
    # kube-master     kube-node-1     kube-node-2    kube-node-extra
    #  10.192.0.2      10.192.0.3      10.192.0.4      10.192.0.5
    #        |                |         | |    |         |
    #        |                +---------+ |    +---------+
    #        +----------------------------+   Peering -> is configured
    #           These peerings are            by get_extra_peer_spec().
    #           configured by BGPPeer         Peering <- is configured
    #           peer-with-rr                  in get_bird_conf().

    def get_bird_conf(self):
        return bird_conf_rr % self.ips[2]

    def get_extra_peer_spec(self):
        return """
spec:
  node: %s
  peerIP: %s
  asNumber: 64512
""" % (self.nodes[2], self.external_node_ip)

    def test_rr(self):
        # Create ExternalTrafficPolicy Local service with one endpoint on node-1
        kubectl("""apply -f - << EOF
apiVersion: apps/v1
kind: Deployment
metadata:
  name: nginx-rr
  namespace: bgp-test
  labels:
    app: nginx
spec:
  replicas: 1
  selector:
    matchLabels:
      app: nginx
      run: nginx-rr
  template:
    metadata:
      labels:
        app: nginx
        run: nginx-rr
    spec:
      containers:
      - name: nginx-rr
        image: %s
        ports:
        - containerPort: 80
      nodeSelector:
        kubernetes.io/os: linux
        kubernetes.io/hostname: %s
---
apiVersion: v1
kind: Service
metadata:
  name: nginx-rr
  namespace: bgp-test
  labels:
    app: nginx
    run: nginx-rr
spec:
  externalIPs:
  - 175.200.1.1
  ports:
  - port: 80
    targetPort: 80
  selector:
    app: nginx
    run: nginx-rr
  type: NodePort
  externalTrafficPolicy: Local
EOF
""" % (NGINX_IMAGE, self.nodes[1]))

        calicoctl("get nodes -o yaml")
        calicoctl("get bgppeers -o yaml")
        calicoctl("get bgpconfigs -o yaml")

        # Update the node-2 to behave as a route-reflector, using a retry
        # to avoid transient conflict errors.
        retry_until_success(lambda: self.set_rr_config(self.nodes[2]))

        # Disable node-to-node mesh, add cluster and external IP CIDRs to
        # advertise, and configure BGP peering between the cluster nodes and the
        # RR.  (The BGP peering from the external node to the RR is included in
        # bird_conf_rr above.)
        calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata:
  name: default
spec:
  nodeToNodeMeshEnabled: false
  asNumber: 64512
  serviceClusterIPs:
  - cidr: 10.96.0.0/12
  serviceExternalIPs:
  - cidr: 175.200.0.0/16
EOF
""")

        calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPPeer
metadata: {name: peer-with-rr}
spec:
  peerIP: %s
  asNumber: 64512
EOF
""" % self.ips[2])
        svc_json = kubectl("get svc nginx-rr -n bgp-test -o json")
        svc_dict = json.loads(svc_json)
        cluster_ip = svc_dict['spec']['clusterIP']
        external_ip = svc_dict['spec']['externalIPs'][0]
        retry_until_success(lambda: self.assertIn(cluster_ip, self.get_routes()))
        retry_until_success(lambda: self.assertIn(external_ip, self.get_routes()))

    def test_single_ip_lb_rr(self):
        """
        Tests a /32 LB service with externalTrafficPolicy=Local using a RR
        """
        # Create an LB ExternalTrafficPolicy Local service with one endpoint
        # on node-1
        kubectl("""apply -f - << EOF
apiVersion: apps/v1
kind: Deployment
metadata:
  name: nginx-rr
  namespace: bgp-test
  labels:
    app: nginx
spec:
  replicas: 1
  selector:
    matchLabels:
      app: nginx
      run: nginx-rr
  template:
    metadata:
      labels:
        app: nginx
        run: nginx-rr
    spec:
      containers:
      - name: nginx-rr
        image: %s
        ports:
        - containerPort: 80
      nodeSelector:
        kubernetes.io/os: linux
        kubernetes.io/hostname: %s
---
apiVersion: projectcalico.org/v3
kind: IPPool
metadata:
 name: loadbalancer-ip-pool
spec:
 cidr: 80.15.0.100/32
 blockSize: 32
 natOutgoing: true
 disabled: false
 assignmentMode: Automatic
 allowedUses:
  - LoadBalancer
---
apiVersion: v1
kind: Service
metadata:
  name: nginx-rr
  namespace: bgp-test
  labels:
    app: nginx
    run: nginx-rr
  annotations:
    projectcalico.org/loadBalancerIPs: '["80.15.0.100"]'
spec:
  ports:
  - port: 80
    targetPort: 80
  selector:
    app: nginx
    run: nginx-rr
  type: LoadBalancer
  loadBalancerIP: 80.15.0.100
  externalTrafficPolicy: Local
EOF
""" % (NGINX_IMAGE, self.nodes[1]))

        calicoctl("get nodes -o yaml")
        calicoctl("get bgppeers -o yaml")
        calicoctl("get bgpconfigs -o yaml")

        # Update the node-2 to behave as a route-reflector, using a retry
        # to avoid transient conflict errors.
        retry_until_success(lambda: self.set_rr_config(self.nodes[2]))

        # Disable node-to-node mesh, add cluster and external IP CIDRs to
        # advertise, and configure BGP peering between the cluster nodes and the
        # RR.  (The BGP peering from the external node to the RR is included in
        # bird_conf_rr above.)
        calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata:
  name: default
spec:
  nodeToNodeMeshEnabled: false
  asNumber: 64512
  serviceClusterIPs:
  - cidr: 10.96.0.0/12
  serviceLoadBalancerIPs:
  - cidr: 80.15.0.100/32
EOF
""")

        calicoctl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPPeer
metadata: {name: peer-with-rr}
spec:
  peerIP: %s
  asNumber: 64512
EOF
""" % self.ips[2])
        svc_json = kubectl("get svc nginx-rr -n bgp-test -o json")
        svc_dict = json.loads(svc_json)
        cluster_ip = svc_dict['spec']['clusterIP']
        load_balancer_ip = svc_dict['spec']['loadBalancerIP']
        retry_until_success(lambda: self.assertIn(cluster_ip, self.get_routes()))
        retry_until_success(lambda: self.assertIn(load_balancer_ip, self.get_routes()))
        kubectl("delete ippool loadbalancer-ip-pool")



# Copyright (c) 2018-2025 Tigera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import re
from subprocess import CalledProcessError

from tests.k8st.test_base import TestBase
from tests.k8st.utils.utils import DiagsCollector, calicoctl, kubectl, exec_in_calico_node, run, retry_until_success, node_info, start_external_node_with_bgp

_log = logging.getLogger(__name__)
_log.setLevel(logging.DEBUG)


class TestReadiness(TestBase):
    def setUp(self):
        # Wait for all Calico pods to be ready.
        kubectl("wait --for=condition=Ready pods --all -n calico-system --timeout=120s")

    def tearDown(self):
        # Make sure nodes are healthy for next test.
        nodes, _, _ = node_info()
        for node in nodes:
            exec_in_calico_node(node, "sv start /etc/service/enabled/bird")
            exec_in_calico_node(node, "sv start /etc/service/enabled/bird6")
            exec_in_calico_node(node, "sv start /etc/service/enabled/confd")
            exec_in_calico_node(node, "sv start /etc/service/enabled/felix")

    def assert_readiness(self, node, flag, ready):
        if not ready:
            try:
                exec_in_calico_node(node, "/bin/calico-node -%s-ready" % flag)
            except CalledProcessError as e:
                _log.info("Got expected error: %s", e)
                return
            raise AssertionError("Expected %s not to be ready" % flag)

        exec_in_calico_node(node, "/bin/calico-node -%s-ready" % flag)

    def assert_liveness(self, node, flag, live):
        if not live:
            try:
                exec_in_calico_node(node, "/bin/calico-node -%s-live" % flag)
            except CalledProcessError as e:
                _log.info("Got expected error: %s", e)
                return
            raise AssertionError("Expected %s not to be live" % flag)

        exec_in_calico_node(node, "/bin/calico-node -%s-live" % flag)

    def test_readiness_bird_down(self):
        """
        Simulate bird service to be down.
        """
        nodes, _, _ = node_info()
        self.assert_readiness(nodes[0], "bird", True)
        exec_in_calico_node(nodes[0], "sv stop /etc/service/enabled/bird")
        self.assert_readiness(nodes[0], "bird", False)
        exec_in_calico_node(nodes[0], "sv start /etc/service/enabled/bird")

    def test_readiness_bird6_down(self):
        """
        Simulate bird6 service to be down.
        """
        nodes, _, _ = node_info()
        self.assert_readiness(nodes[0], "bird6", True)
        exec_in_calico_node(nodes[0], "sv stop /etc/service/enabled/bird6")
        self.assert_readiness(nodes[0], "bird6", False)
        exec_in_calico_node(nodes[0], "sv start /etc/service/enabled/bird6")

    def test_readiness_confd_down(self):
        """
        Simulate confd service to be down.
        """
        nodes, _, _ = node_info()
        self.assert_liveness(nodes[0], "bird", True)
        exec_in_calico_node(nodes[0], "sv stop /etc/service/enabled/confd")
        self.assert_liveness(nodes[0], "bird", False)
        exec_in_calico_node(nodes[0], "sv start /etc/service/enabled/confd")

    def test_felix_down(self):
        """
        Simulate felix service to be down.
        """
        nodes, _, _ = node_info()
        self.assert_readiness(nodes[0], "felix", True)
        exec_in_calico_node(nodes[0], "sv stop /etc/service/enabled/felix")
        self.assert_readiness(nodes[0], "felix", False)
        exec_in_calico_node(nodes[0], "sv start /etc/service/enabled/felix")



# Copyright (c) 2018 Tigera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import functools
import json
import logging
import os
import random
import string
import subprocess
import time

from kubernetes import client, config

_log = logging.getLogger(__name__)

ROUTER_IMAGE = os.getenv("ROUTER_IMAGE", "calico/bird:latest")
NGINX_IMAGE = os.getenv("NGINX_IMAGE", "nginx:1")


# Helps with printing diags after a test.
class DiagsCollector(object):
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        # Print out diagnostics for the test. These will go to screen
        # on test failure.
        _log.info("===================================================")
        _log.info("============= COLLECTING DIAGS FOR TEST ===========")
        _log.info("===================================================")
        kubectl("version")
        kubectl("get deployments,pods,svc,endpoints --all-namespaces -o wide")
        for resource in ["node", "bgpconfig", "bgppeer", "gnp", "felixconfig"]:
            _log.info("")
            calicoctl("get " + resource + " -o yaml")
        nodes, _, _ = node_info()
        for node in nodes:
            _log.info("")
            run("docker exec " + node + " ip r")
            run("docker exec " + node + " ip -6 r")
        kubectl("logs -n calico-system -l k8s-app=calico-node")
        self.print_confd_templates(nodes)
        _log.info("===================================================")
        _log.info("============= COLLECTED DIAGS FOR TEST ============")
        _log.info("===================================================")

    def print_confd_templates(self, nodes):
        for node in nodes:
            calicoPod = kubectl("-n calico-system get pods -o wide | grep calico-node | grep '%s '| cut -d' ' -f1" % node)
            if calicoPod is None:
                continue
            calicoPod = calicoPod.strip()

            # v4 files.
            kubectl("exec -n calico-system %s -- cat /etc/calico/confd/config/bird.cfg" % calicoPod)
            kubectl("exec -n calico-system %s -- cat /etc/calico/confd/config/bird_aggr.cfg" % calicoPod)
            kubectl("exec -n calico-system %s -- cat /etc/calico/confd/config/bird_ipam.cfg" % calicoPod)

            # And for v6.
            kubectl("exec -n calico-system %s -- cat /etc/calico/confd/config/bird6.cfg" % calicoPod)
            kubectl("exec -n calico-system %s -- cat /etc/calico/confd/config/bird6_aggr.cfg" % calicoPod)
            kubectl("exec -n calico-system %s -- cat /etc/calico/confd/config/bird6_ipam.cfg" % calicoPod)

def exec_in_calico_node(node, command):
    calicoPod = kubectl("-n calico-system get pods -o wide | grep calico-node | grep '%s '| cut -d' ' -f1" % node)
    if calicoPod is None:
        raise Exception("No calico-node pod found on node %s" % node)
    calicoPod = calicoPod.strip()
    return kubectl("exec -n calico-system %s -- %s" % (calicoPod, command))

def start_external_node_with_bgp(name, bird_peer_config=None, bird6_peer_config=None):
    # Check how much disk space we have.
    run("df -h")

    # Setup external node: use privileged mode for setting routes.
    run("docker run -d --privileged --net=kind --name %s %s" % (name, ROUTER_IMAGE))

    # Check how much space there is inside the container.  We may need
    # to retry this, as it may take a while for the image to download
    # and the container to start running.
    while True:
        try:
            run("docker exec %s df -h" % name)
            break
        except subprocess.CalledProcessError:
            _log.exception("Container not ready yet")
            time.sleep(20)

    # Install curl and iproute2.
    run("docker exec %s apk add --no-cache curl iproute2" % name)

    # Set ECMP hash algorithm to L4 for a proper load balancing between nodes.
    run("docker exec %s sysctl -w net.ipv4.fib_multipath_hash_policy=1" % name)

    # Add "merge paths on" to the BIRD config.
    run("docker exec %s sed -i '/protocol kernel {/a merge paths on;' /etc/bird.conf" % name)
    run("docker exec %s sed -i '/protocol kernel {/a merge paths on;' /etc/bird6.conf" % name)

    if bird_peer_config:
        # Install desired peer config.
        output = run("docker inspect -f '{{range .NetworkSettings.Networks}}{{.IPAddress}}{{end}}' %s" % name)
        birdy_ip = output.strip()
        with open('peers.conf', 'w') as peerconfig:
            peerconfig.write(bird_peer_config.replace("ip@local", birdy_ip))
        run("docker cp peers.conf %s:/etc/bird/peers.conf" % name)
        run("rm peers.conf")
        run("docker exec %s birdcl configure" % name)

    elif bird6_peer_config:
        # Install desired peer config.
        birdy_ip = "2001:20::20"
        run("docker exec %s sysctl -w net.ipv6.conf.all.disable_ipv6=0" % name)
        run("docker exec %s sysctl -w net.ipv6.conf.all.forwarding=1" % name)

        # Try to set net.ipv6.fib_multipath_hash_policy to get IPv6
        # ECMP load balancing by 5-tuple, but allow it to fail as
        # older kernels (e.g. Semaphore v2) don't have that setting.
        # It doesn't actually matter as we aren't currently testing
        # IPv6 ECMP behaviour in detail.
        run("docker exec %s sysctl -w net.ipv6.fib_multipath_hash_policy=1" % name,
            allow_fail=True)

        run("docker exec %s ip -6 a a %s/64 dev eth0" % (name, birdy_ip))
        with open('peers.conf', 'w') as peerconfig:
            peerconfig.write(bird6_peer_config.replace("ip@local", birdy_ip))
        run("docker cp peers.conf %s:/etc/bird6/peers.conf" % name)
        run("rm peers.conf")
        run("docker exec %s birdcl6 configure" % name)

    return birdy_ip

def retry_until_success(fun,
                        retries=90,
                        wait_time=1,
                        ex_class=None,
                        log_exception=True,
                        function_args=None,
                        function_kwargs=None):
    """
    Retries function until no exception is thrown. If exception continues,
    it is reraised.
    :param fun: the function to be repeatedly called
    :param retries: the maximum number of times to retry the function.  A value
    of 0 will run the function once with no retries.
    :param wait_time: the time to wait between retries (in s)
    :param ex_class: The class of expected exceptions.
    :param log_exception: By default this function logs the exception if the
    function is still failing after max retries.   This log can sometimes be
    superfluous -- if e.g. the calling code is going to make a better log --
    and can be suppressed by setting this parameter to False.
    :param function_args: A list of arguments to pass to function
    :param function_kwargs: A dictionary of keyword arguments to pass to
                            function
    :returns: the value returned by function
    """
    if function_args is None:
        function_args = []
    if function_kwargs is None:
        function_kwargs = {}
    for retry in range(retries + 1):
        try:
            result = fun(*function_args, **function_kwargs)
        except Exception as e:
            if ex_class and e.__class__ is not ex_class:
                _log.exception("Hit unexpected exception in function - "
                               "not retrying.")
                raise
            if retry < retries:
                _log.debug("Hit exception in function - retrying: %s", e)
                time.sleep(wait_time)
            else:
                if log_exception:
                    _log.exception("Function %s did not succeed before "
                                   "timeout.", fun)
                raise
        else:
            # Successfully ran the function
            return result


def function_name(f):
    """
    A function that returns the name of the provided function as a string.
    This primarily exists to handle the fact that functools.partial is an
    imperfect wrapper.
    """
    if isinstance(f, functools.partial):
        f = f.func

    try:
        return f.__name__
    except Exception:
        return "<unknown function>"


def run(command, logerr=True, allow_fail=False, allow_codes=[], returnerr=False):
    out = ""
    _log.info("[%s] %s", datetime.datetime.now(), command)

    process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = process.communicate()
    out = out.decode()
    err = err.decode()
    _log.info("Out:\n%s", out)
    _log.info("Err:\n%s", err)

    retcode = process.poll()
    if retcode:
        if logerr:
            _log.exception("Failure output:\n%s\nerr:\n%s", out, err)
        if not allow_fail:
            raise subprocess.CalledProcessError(retcode, command, output="stdout: " + out + " stderr: " + err)
        if returnerr:
            return err
    return out


def curl(hostname, container="kube-node-extra"):
    if ':' in hostname:
        # It's an IPv6 address.
        hostname = '[' + hostname + ']'

    cmd = "docker exec %s curl --connect-timeout 2 -m 3 %s" % (container,
                                                               hostname)
    return run(cmd)


def kubectl(args, logerr=True, allow_fail=False, allow_codes=[], timeout=0, returnerr=False):
    if timeout == 0:
        cmd = "kubectl "
    else:
        cmd = "timeout -s %d kubectl " % timeout
    return run(cmd + args,
               logerr=logerr,
               allow_fail=allow_fail,
               allow_codes=allow_codes,
               returnerr=returnerr)

def calicoctl(args, allow_fail=False):
    return kubectl("exec -i -n kube-system calicoctl -- calicoctl --allow-version-mismatch " + args,
                   allow_fail=allow_fail)


def calicoctl_apply_dict(object_dict):
    calicoctl("""apply -f - << EOF
%s
EOF
""" % json.dumps(object_dict))


def generate_unique_id(length, prefix=""):
    random_string = ''.join(random.choice(string.ascii_lowercase + string.digits) for _ in range(length))
    return "%s-%s" % (prefix, random_string)


# We have to define and use this static map, from each node name to
# its IPv6 address, because Kubernetes does not yet allow for an IPv6
# address field in its host resource.  The mappings here must match
# the code in tests/k8st/deploy_resources_on_kind_cluster.sh that assigns an IPv6
# address to each node.
ipv6_map = {
    "kind-control-plane": "2001:20::8",
    "kind-worker": "2001:20::1",
    "kind-worker2": "2001:20::2",
    "kind-worker3": "2001:20::3",
}


def node_info():
    nodes = []
    ips = []
    ip6s = []

    master_node = kubectl("get node --selector='node-role.kubernetes.io/control-plane' -o jsonpath='{.items[0].metadata.name}'")
    nodes.append(master_node)
    ip6s.append(ipv6_map[master_node])
    master_ip = kubectl("get node --selector='node-role.kubernetes.io/control-plane' -o jsonpath='{.items[0].status.addresses[0].address}'")
    ips.append(master_ip)

    for i in range(3):
        node = kubectl("get node --selector='!node-role.kubernetes.io/control-plane' -o jsonpath='{.items[%d].metadata.name}'" % i)
        nodes.append(node)
        ip6s.append(ipv6_map[node])
        node_ip = kubectl("get node --selector='!node-role.kubernetes.io/control-plane' -o jsonpath='{.items[%d].status.addresses[0].address}'" % i)
        ips.append(node_ip)
    return nodes, ips, ip6s

def calico_node_pod_name(nodename):
    name = kubectl("get po -n calico-system -l k8s-app=calico-node --field-selector spec.nodeName=%s -o jsonpath='{.items[0].metadata.name}'" % nodename)
    return name



# Copyright 2016 Metaswitch Networks
# Copyright 2018 Tigera, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import logging
import socket
from collections import namedtuple

import eventlet

import mock

from neutron.agent.dhcp_agent import register_options
from neutron.agent.linux import dhcp
from neutron.tests import base

from neutron_lib.constants import DHCPV6_STATEFUL, IPV6_SLAAC

from oslo_config import cfg

from networking_calico import datamodel_v1
from networking_calico import datamodel_v2
from networking_calico.agent.dhcp_agent import CalicoDhcpAgent
from networking_calico.agent.dhcp_agent import FakePlugin
from networking_calico.agent.linux.dhcp import DnsmasqRouted
from networking_calico.common import config as calico_config
from networking_calico.etcdutils import EtcdWatcher

LOG = logging.getLogger(__name__)

EtcdResponse = namedtuple("EtcdResponse", ["value"])


def make_endpoint_name(endpoint_id):
    parts = [socket.gethostname(), "openstack", "workload-ignored", endpoint_id]
    return "-".join([p.replace("-", "--") for p in parts])


class TestFakePlugin(base.BaseTestCase):
    def setUp(self):
        super(TestFakePlugin, self).setUp()
        self.plugin = FakePlugin()

    def test_create(self):
        port = self.plugin.create_dhcp_port({"port": {"network_id": "net-id-0"}})
        self.assertEqual(
            {
                "network_id": "net-id-0",
                "device_owner": "network:dhcp",
                "id": "net-id-0",
                "mac_address": "02:00:00:00:00:00",
            },
            port,
        )

    def test_release(self):
        self.plugin.release_dhcp_port("calico", "dhcp")


class TestDhcpAgent(base.BaseTestCase):
    def setUp(self):
        super(TestDhcpAgent, self).setUp()
        register_options(cfg.CONF)
        calico_config.register_options(cfg.CONF)
        self.mock_makedirs_p = mock.patch("os.makedirs")
        self.mock_makedirs = self.mock_makedirs_p.start()
        self.hostname = socket.gethostname()
        cfg.CONF.host = self.hostname

    def test_mainline(self):
        # Create the DHCP agent.
        agent = CalicoDhcpAgent()

        # Mock the MTU watcher.
        agent.etcd.mtu_watcher = mock.Mock()
        agent.etcd.mtu_watcher.get_mtu.return_value = 1500

        # Mock the dnsmasq updater.
        agent.etcd.dnsmasq_updater = mock.Mock()

        # Check that running it invokes the etcd watcher loop.
        with mock.patch.object(agent, "etcd") as etcdobj:
            agent.run()
            etcdobj.start.assert_called_with()

        # Notify initial snapshot (empty).
        snapshot_data = agent.etcd._pre_snapshot_hook()
        agent.etcd._post_snapshot_hook(snapshot_data)
        agent.etcd.dnsmasq_updater.update_network.assert_not_called()

        # Notify subnets.
        agent.etcd.subnet_watcher.on_subnet_set(
            EtcdResponse(
                value=json.dumps(
                    {
                        "cidr": "10.28.0.0/24",
                        "gateway_ip": "10.28.0.1",
                        "host_routes": [],
                    }
                )
            ),
            "v4subnet-1",
        )
        agent.etcd.subnet_watcher.on_subnet_set(
            EtcdResponse(
                value=json.dumps(
                    {"cidr": "2001:db8:1::/80", "gateway_ip": "2001:db8:1::1"}
                )
            ),
            "v6subnet-1",
        )
        agent.etcd.subnet_watcher.on_subnet_set(
            EtcdResponse(
                value=json.dumps(
                    {
                        "cidr": "10.29.0.0/24",
                        "gateway_ip": "10.29.0.1",
                        "host_routes": [
                            {"destination": "11.11.0.0/16", "nexthop": "10.65.0.1"}
                        ],
                    }
                )
            ),
            "v4subnet-2",
        )

        # Notify an endpoint.
        agent.etcd.on_endpoint_set(
            EtcdResponse(
                value=json.dumps(
                    {
                        "spec": {
                            "interfaceName": "tap1234",
                            "mac": "fe:16:65:12:33:44",
                            "profiles": ["profile-1"],
                            "ipNetworks": ["10.28.0.2/32", "2001:db8:1::2/128"],
                            "ipv4Gateway": "10.28.0.1",
                            "ipv6Gateway": "2001:db8:1::1",
                        }
                    }
                )
            ),
            make_endpoint_name("endpoint-1"),
        )

        # Check DHCP driver was asked to restart.
        agent.etcd.dnsmasq_updater.update_network.assert_called()
        agent.etcd.dnsmasq_updater.update_network.reset_mock()

        # Notify another endpoint (using the same subnets).
        agent.etcd.on_endpoint_set(
            EtcdResponse(
                value=json.dumps(
                    {
                        "spec": {
                            "interfaceName": "tap5678",
                            "mac": "fe:16:65:12:33:55",
                            "profiles": ["profile-1"],
                            "ipNetworks": ["10.28.0.3/32", "2001:db8:1::3/128"],
                            "ipv4Gateway": "10.28.0.1",
                            "ipv6Gateway": "2001:db8:1::1",
                            "fqdn": "calico-vm17.datcon.co.uk",
                        }
                    }
                )
            ),
            make_endpoint_name("endpoint-2"),
        )

        # Check DHCP driver was asked to restart.
        agent.etcd.dnsmasq_updater.update_network.assert_called()
        agent.etcd.dnsmasq_updater.update_network.reset_mock()

        # Notify deletion of the first endpoint.
        agent.etcd.on_endpoint_delete(None, make_endpoint_name("endpoint-1"))

        # Check DHCP driver was asked to reload allocations.
        agent.etcd.dnsmasq_updater.update_network.assert_called()
        agent.etcd.dnsmasq_updater.update_network.reset_mock()

        # Notify another endpoint using a new subnet.
        agent.etcd.on_endpoint_set(
            EtcdResponse(
                value=json.dumps(
                    {
                        "spec": {
                            "interfaceName": "tapABCD",
                            "mac": "fe:16:65:12:33:66",
                            "profiles": ["profile-1"],
                            "ipNetworks": ["10.29.0.3/32"],
                            "ipv4Gateway": "10.29.0.1",
                        }
                    }
                )
            ),
            make_endpoint_name("endpoint-3"),
        )

        # Check DHCP driver was asked to restart.
        agent.etcd.dnsmasq_updater.update_network.assert_called()
        agent.etcd.dnsmasq_updater.update_network.reset_mock()

    def test_initial_snapshot(self):
        # Create the DHCP agent.
        agent = CalicoDhcpAgent()

        # Check that running it invokes the etcd watcher loop.
        with mock.patch.object(agent, "etcd") as etcdobj:
            agent.run()
            etcdobj.start.assert_called_with()

        with mock.patch.object(agent, "call_driver") as call_driver:
            # Notify a non-empty initial snapshot.
            snapshot_data = agent.etcd._pre_snapshot_hook()
            agent.etcd.on_endpoint_set(
                EtcdResponse(
                    value=json.dumps(
                        {
                            "spec": {
                                "interfaceName": "tap1234",
                                "mac": "fe:16:65:12:33:44",
                                "profiles": ["profile-1"],
                                "ipNetworks": ["10.28.0.2/32"],
                                "ipv4Gateway": "10.28.0.1",
                            }
                        }
                    )
                ),
                make_endpoint_name("endpoint-4"),
            )
            agent.etcd._post_snapshot_hook(snapshot_data)

            # Check DHCP driver was not troubled - because the subnet data was
            # missing and so the port could not be processed further.
            call_driver.assert_not_called()

    @mock.patch.object(EtcdWatcher, "start")
    def test_kill_agent(self, loop_fn):

        # To test handling of SubnetWatcher's loop exiting, make
        # EtcdWatcher.start throw an exception.
        loop_fn.side_effect = Exception("from test_kill_agent")

        # Create the DHCP agent and allow it to start the SubnetWatcher loop.
        agent = CalicoDhcpAgent()
        agent.etcd._stopped = False
        eventlet.spawn(agent.etcd.subnet_watcher.start)
        eventlet.sleep(0.2)

        # Check that exception handling caused the endpoint watcher to be
        # marked as stopped.
        self.assertTrue(agent.etcd._stopped)

    def test_invalid_endpoint_data(self):
        # Create the DHCP agent.
        agent = CalicoDhcpAgent()

        with mock.patch.object(agent, "call_driver") as call_driver:
            # Notify an endpoint missing some required fields.
            agent.etcd.on_endpoint_set(
                EtcdResponse(
                    value=json.dumps(
                        {"mac": "fe:16:65:12:33:44", "ipNetworks": ["2001:db8:1::1"]}
                    )
                ),
                make_endpoint_name("endpoint-1"),
            )

            # Check DHCP driver was not asked to do anything.
            call_driver.assert_not_called()

            # Notify an endpoint with non-dict data.
            agent.etcd.on_endpoint_set(
                EtcdResponse(value="not even a dict!"), make_endpoint_name("endpoint-1")
            )

            # Check DHCP driver was not asked to do anything.
            call_driver.assert_not_called()

            # One more variant.
            agent.etcd.on_endpoint_set(
                EtcdResponse(value='"nor this!"'), make_endpoint_name("endpoint-1")
            )

            # Check DHCP driver was not asked to do anything.
            call_driver.assert_not_called()

    def test_endpoint_no_ipnetworks(self):
        # Create the DHCP agent.
        agent = CalicoDhcpAgent()

        with mock.patch.object(agent, "call_driver") as call_driver:
            with mock.patch.object(agent.etcd, "on_endpoint_delete") as ep_del:
                # Notify an endpoint that is valid but has no ipNetworks.
                agent.etcd.on_endpoint_set(
                    EtcdResponse(
                        value=json.dumps(
                            {
                                "spec": {
                                    "interfaceName": "tapfe166512-33",
                                    "mac": "fe:16:65:12:33:44",
                                    "ipNetworks": [],
                                }
                            }
                        )
                    ),
                    make_endpoint_name("endpoint-1"),
                )

                # Check handled as a deletion.
                ep_del.assert_called()

            # Check DHCP driver was not asked to do anything.
            call_driver.assert_not_called()

    def test_no_subnet_data(self):
        # Create the DHCP agent.
        agent = CalicoDhcpAgent()

        with mock.patch.object(agent, "call_driver") as call_driver:
            # Notify an endpoint.
            agent.etcd.on_endpoint_set(
                EtcdResponse(
                    value=json.dumps(
                        {
                            "spec": {
                                "interfaceName": "tap1234",
                                "mac": "fe:16:65:12:33:44",
                                "profiles": ["profile-1"],
                                "ipNetworks": ["10.28.0.2/32", "2001:db8:1::2/128"],
                                "ipv4Gateway": "10.28.0.1",
                                "ipv6Gateway": "2001:db8:1::1",
                            }
                        }
                    )
                ),
                make_endpoint_name("endpoint-1"),
            )
            call_driver.assert_not_called()

    def test_invalid_subnet_data(self):
        # Create the DHCP agent.
        agent = CalicoDhcpAgent()

        agent.etcd.subnet_watcher.on_subnet_set(
            EtcdResponse(value=json.dumps({"gateway_ip": "10.28.0.1"})), "v4subnet-1"
        )

        agent.etcd.subnet_watcher.on_subnet_set(
            EtcdResponse(value=json.dumps({"gateway_ip": "2001:db8:1::1"})),
            "v6subnet-1",
        )

        self.assertFalse(agent.etcd.subnet_watcher.subnets_by_id)

    def test_etcd_watchers_init_with_conf_values(self):
        agent = CalicoDhcpAgent()
        self.assertEqual(
            agent.etcd.prefix,
            "/calico/resources/v3/projectcalico.org/"
            + "workloadendpoints/openstack/"
            + self.hostname.replace("-", "--")
            + "-openstack-",
        )
        self.assertEqual(agent.etcd.v1_subnet_watcher.prefix, datamodel_v1.SUBNET_DIR)
        self.assertEqual(agent.etcd.subnet_watcher.prefix, datamodel_v2.subnet_dir())

    def test_host_config(self):
        cfg.CONF.host = "my-special-hostname"
        agent = CalicoDhcpAgent()
        self.assertEqual(
            agent.etcd.prefix,
            "/calico/resources/v3/projectcalico.org/"
            + "workloadendpoints/openstack/"
            + "my--special--hostname"
            + "-openstack-",
        )
        self.assertEqual(agent.etcd.v1_subnet_watcher.prefix, datamodel_v1.SUBNET_DIR)
        self.assertEqual(agent.etcd.subnet_watcher.prefix, datamodel_v2.subnet_dir())

    def test_region_config(self):
        cfg.CONF.set_override("openstack_region", "asia-central", group="calico")
        calico_config._reset_globals()
        agent = CalicoDhcpAgent()
        self.assertEqual(
            agent.etcd.prefix,
            "/calico/resources/v3/projectcalico.org/"
            + "workloadendpoints/openstack-region-asia-central/"
            + self.hostname.replace("-", "--")
            + "-openstack-",
        )
        self.assertEqual(agent.etcd.v1_subnet_watcher.prefix, datamodel_v1.SUBNET_DIR)
        self.assertEqual(
            agent.etcd.subnet_watcher.prefix,
            datamodel_v2.subnet_dir("region-asia-central"),
        )
        calico_config._reset_globals()


class TestDnsmasqRouted(base.BaseTestCase):
    def setUp(self):
        super(TestDnsmasqRouted, self).setUp()
        register_options(cfg.CONF)
        cfg.CONF.set_override("dhcp_confs", "/run")
        cfg.CONF.set_override(
            "interface_driver",
            "networking_calico.agent.linux.interface.RoutedInterfaceDriver",
        )
        self.mock_makedirs_p = mock.patch("os.makedirs")
        self.mock_makedirs = self.mock_makedirs_p.start()

    @mock.patch("neutron.agent.linux.dhcp.DeviceManager")
    @mock.patch("neutron.agent.linux.dhcp.common_utils")
    def test_build_cmdline(self, commonutils, device_mgr_cls):
        v4subnet = mock.Mock()
        v4subnet.id = "v4subnet-1"
        v4subnet.enable_dhcp = True
        v4subnet.ip_version = 4
        v4subnet.cidr = "10.28.0.0/24"
        v6subnet = mock.Mock()
        v6subnet.id = "v6subnet-1"
        v6subnet.enable_dhcp = True
        v6subnet.ip_version = 6
        v6subnet.cidr = "2001:db8:1::/80"
        v6subnet.ipv6_ra_mode = DHCPV6_STATEFUL
        v6subnet.ipv6_address_mode = DHCPV6_STATEFUL
        network = mock.Mock()
        network.id = "calico"
        network.subnets = [v4subnet, v6subnet]
        network.mtu = 0
        network.ports = [
            dhcp.DictModel({"device_id": "tap1"}),
            dhcp.DictModel({"device_id": "tap2"}),
            dhcp.DictModel({"device_id": "tap3"}),
        ]
        network.non_local_subnets = []
        network.get.side_effect = lambda key, dflt=None: dflt
        device_mgr_cls.return_value.driver.bridged = False
        dhcp_driver = DnsmasqRouted(cfg.CONF, network, None, plugin=FakePlugin())
        with mock.patch.object(dhcp_driver, "_get_value_from_conf_file") as gv:
            gv.return_value = "ns-dhcp"
            cmdline = dhcp_driver._build_cmdline_callback("/run/pid_file")

        # Filter out dnsmasq args that we don't care about.
        filtered_args = []
        for arg in cmdline:
            if "--domain=" in arg:
                continue
            if arg in [
                "--no-hosts",
                "--no-resolv",
                "--pid-file=/run/pid_file",
                "--dhcp-hostsfile=/run/calico/host",
                "--addn-hosts=/run/calico/addn_hosts",
                "--dhcp-optsfile=/run/calico/opts",
                "--dhcp-leasefile=/run/calico/leases",
                "--dhcp-match=set:ipxe,175",
                "--dhcp-lease-max=16777216",
                "--conf-file=",
            ]:
                continue
            filtered_args.append(arg)

        # Check the remaining filtered args against what we expect.
        self.assertEqual(
            [
                "dnsmasq",
                "--except-interface=lo",
                "--bind-dynamic",
                "--interface=ns-dhcp",
                "--dhcp-range=set:subnet-v4subnet-1,10.28.0.0"
                + ",static,255.255.255.0,86400s",
                "--dhcp-range=set:subnet-v6subnet-1,2001:db8:1::"
                + ",static,off-link,80,86400s",
                "--enable-ra",
                "--ra-param=ns-*,mtu:off,0",
                "--interface=tap1",
                "--interface=tap2",
                "--interface=tap3",
                "--bridge-interface=ns-dhcp,tap1,tap2,tap3",
            ],
            filtered_args,
        )

    @mock.patch("neutron.agent.linux.dhcp.DeviceManager")
    @mock.patch("neutron.agent.linux.dhcp.common_utils")
    def test_build_cmdline_slaac(self, commonutils, device_mgr_cls):
        v4subnet = mock.Mock()
        v4subnet.id = "v4subnet-1"
        v4subnet.enable_dhcp = True
        v4subnet.ip_version = 4
        v4subnet.cidr = "10.28.0.0/24"
        v6subnet = mock.Mock()
        v6subnet.id = "v6subnet-1"
        v6subnet.enable_dhcp = True
        v6subnet.ip_version = 6
        v6subnet.cidr = "2001:db8:1::/80"
        v6subnet.ipv6_ra_mode = DHCPV6_STATEFUL
        v6subnet.ipv6_address_mode = DHCPV6_STATEFUL
        v6subnet_slaac = mock.Mock()
        v6subnet_slaac.id = "v6subnet-2"
        v6subnet_slaac.enable_dhcp = True
        v6subnet_slaac.ip_version = 6
        v6subnet_slaac.cidr = "2001:e000:1::/64"
        v6subnet_slaac.ipv6_ra_mode = IPV6_SLAAC
        v6subnet_slaac.ipv6_address_mode = IPV6_SLAAC
        network = mock.Mock()
        network.id = "calico"
        network.subnets = [v4subnet, v6subnet, v6subnet_slaac]
        network.mtu = 0
        network.ports = [
            dhcp.DictModel({"device_id": "tap1"}),
            dhcp.DictModel({"device_id": "tap2"}),
            dhcp.DictModel({"device_id": "tap3"}),
        ]
        network.non_local_subnets = []
        network.get.side_effect = lambda key, dflt=None: dflt
        device_mgr_cls.return_value.driver.bridged = False
        dhcp_driver = DnsmasqRouted(cfg.CONF, network, None, plugin=FakePlugin())
        with mock.patch.object(dhcp_driver, "_get_value_from_conf_file") as gv:
            gv.return_value = "ns-dhcp"
            cmdline = dhcp_driver._build_cmdline_callback("/run/pid_file")

        # Filter out dnsmasq args that we don't care about.
        filtered_args = []
        for arg in cmdline:
            if "--domain=" in arg:
                continue
            if arg in [
                "--no-hosts",
                "--no-resolv",
                "--pid-file=/run/pid_file",
                "--dhcp-hostsfile=/run/calico/host",
                "--addn-hosts=/run/calico/addn_hosts",
                "--dhcp-optsfile=/run/calico/opts",
                "--dhcp-leasefile=/run/calico/leases",
                "--dhcp-match=set:ipxe,175",
                "--dhcp-lease-max=16777216",
                "--conf-file=",
            ]:
                continue
            filtered_args.append(arg)

        # Check the remaining filtered args against what we expect.
        self.assertEqual(
            [
                "dnsmasq",
                "--except-interface=lo",
                "--bind-dynamic",
                "--interface=ns-dhcp",
                "--dhcp-range=set:subnet-v4subnet-1,10.28.0.0"
                + ",static,255.255.255.0,86400s",
                "--dhcp-range=set:subnet-v6subnet-2,2001:e000:1::"
                + ",slaac,ra-only,64,86400s",
                "--dhcp-range=set:subnet-v6subnet-1,2001:db8:1::"
                + ",static,off-link,80,86400s",
                "--enable-ra",
                "--ra-param=ns-*,mtu:off,0",
                "--interface=tap1",
                "--interface=tap2",
                "--interface=tap3",
                "--bridge-interface=ns-dhcp,tap1,tap2,tap3",
            ],
            filtered_args,
        )



# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# THIS FILE IS MANAGED BY THE GLOBAL REQUIREMENTS REPO - DO NOT EDIT
from setuptools import find_packages
from setuptools import setup

setup(
    name="networking-calico",
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "calico-dhcp-agent = networking_calico.agent.dhcp_agent:main",
        ],
        "neutron.ml2.mechanism_drivers": [
            (
                "calico = networking_calico.plugins.ml2.drivers.calico."
                "mech_calico:CalicoMechanismDriver"
            ),
        ],
        "neutron.core_plugins": [
            "calico = networking_calico.plugins.calico.plugin:CalicoPlugin",
        ],
    },
    version="0.0.0",
)



# Copyright (c) 2018 Tigera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
import os
import subprocess
import time
from pprint import pformat
from unittest import TestCase

from deepdiff import DeepDiff
from kubernetes import client, config

from tests.k8st.utils.utils import retry_until_success, run, kubectl

logger = logging.getLogger(__name__)


first_log_time = None


class TestBase(TestCase):

    """
    Base class for test-wide methods.
    """

    def setUp(self):
        """
        Clean up before every test.
        """
        self.cluster = self.k8s_client()
        self.cleanups = []

        # Log a newline to ensure that the first log appears on its own line.
        logger.info("")

    def tearDown(self):
        errors = []
        for cleanup in reversed(self.cleanups):
            try:
                cleanup()
            except Exception as e:
                logger.error("Error during cleanup: %s", e)
                errors.append(e)
        super(TestBase, self).tearDown()
        if errors:
            raise Exception("Errors during cleanup: %s" % errors)

    def add_cleanup(self, cleanup):
        self.cleanups.append(cleanup)

    @staticmethod
    def assert_same(thing1, thing2):
        """
        Compares two things.  Debug logs the differences between them before
        asserting that they are the same.
        """
        assert thing1 == thing2, \
            "Items are not the same.  Difference is:\n %s" % \
            pformat(DeepDiff(thing1, thing2), indent=2)

    @staticmethod
    def writejson(filename, data):
        """
        Converts a python dict to json and outputs to a file.
        :param filename: filename to write
        :param data: dictionary to write out as json
        """
        with open(filename, 'w') as f:
            text = json.dumps(data,
                              sort_keys=True,
                              indent=2,
                              separators=(',', ': '))
            logger.debug("Writing %s: \n%s" % (filename, text))
            f.write(text)

    @staticmethod
    def log_banner(msg, *args, **kwargs):
        global first_log_time
        time_now = time.time()
        if first_log_time is None:
            first_log_time = time_now
        time_now -= first_log_time
        elapsed_hms = "%02d:%02d:%02d " % (time_now // 3600,
                                           (time_now % 3600) // 60,
                                           time_now % 60)

        level = kwargs.pop("level", logging.INFO)
        msg = elapsed_hms + str(msg) % args
        banner = "+" + ("-" * (len(msg) + 2)) + "+"
        logger.log(level, "\n" +
                   banner + "\n"
                            "| " + msg + " |\n" +
                   banner)

    @staticmethod
    def k8s_client():
        config.load_kube_config(os.environ.get('KUBECONFIG'))
        return client.CoreV1Api()

    def check_pod_status(self, ns):
        pods = self.cluster.list_namespaced_pod(ns)

        for pod in pods.items:
            logger.info("%s\t%s\t%s", pod.metadata.name, pod.metadata.namespace, pod.status.phase)
            if pod.status.phase != 'Running':
                kubectl("describe po %s -n %s" % (pod.metadata.name, pod.metadata.namespace))
            assert pod.status.phase == 'Running'

    def create_namespace(self, ns_name):
        self.cluster.create_namespace(client.V1Namespace(metadata=client.V1ObjectMeta(name=ns_name)))

    def deploy(self, image, name, ns, port, replicas=1, svc_type="NodePort", traffic_policy="Local", cluster_ip=None, ext_ip=None, ipv6=False):
        """
        Creates a deployment and corresponding service with the given
        parameters.
        """
        # Use a pod anti-affinity so that the scheduler prefers deploying the
        # pods on different nodes. This makes our tests more reliable, since
        # some tests expect pods to be scheduled to different nodes.
        selector = {'matchLabels': {'app': name}}
        terms = [client.V1WeightedPodAffinityTerm(
            pod_affinity_term=client.V1PodAffinityTerm(
                label_selector=selector,
                topology_key="kubernetes.io/hostname"),
            weight=100,
            )]
        anti_aff = client.V1PodAntiAffinity(
                preferred_during_scheduling_ignored_during_execution=terms)

        # Run a deployment with <replicas> copies of <image>, with the
        # pods labelled with "app": <name>.
        deployment = client.V1Deployment(
            api_version="apps/v1",
            kind="Deployment",
            metadata=client.V1ObjectMeta(name=name),
            spec=client.V1DeploymentSpec(
                replicas=replicas,
                selector=selector,
                template=client.V1PodTemplateSpec(
                    metadata=client.V1ObjectMeta(labels={"app": name}),
                    spec=client.V1PodSpec(
                        affinity=client.V1Affinity(pod_anti_affinity=anti_aff),
                        containers=[
                          client.V1Container(name=name,
                                             image=image,
                                             ports=[client.V1ContainerPort(container_port=port)]),
                    ]))))

        # Create the deployment.
        api_response = client.AppsV1Api().create_namespaced_deployment(
            body=deployment,
            namespace=ns)
        logger.debug("Deployment created. status='%s'" % str(api_response.status))

        # Create a service called <name> whose endpoints are the pods
        # with "app": <name>; i.e. those just created above.
        self.create_service(name, name, ns, port, svc_type, traffic_policy, ext_ip=ext_ip, ipv6=ipv6)

    def wait_for_deployment(self, name, ns):
        """
        Waits for the given deployment to have the desired number of replicas.
        """
        logger.info("Checking status for deployment %s/%s" % (ns, name))
        kubectl("-n %s rollout status deployment/%s" % (ns, name))
        kubectl("get pods -n %s -o wide" % ns)

    def create_service(self, name, app, ns, port, svc_type="NodePort", traffic_policy="Local", cluster_ip=None, ext_ip=None, ipv6=False):
        service = client.V1Service(
            metadata=client.V1ObjectMeta(
                name=name,
                labels={"name": name},
            ),
            spec={
                "ports": [{"port": port}],
                "selector": {"app": app},
                "type": svc_type,
                "externalTrafficPolicy": traffic_policy,
            }
        )
        if cluster_ip:
          service.spec["clusterIP"] = cluster_ip
        if ext_ip:
          service.spec["externalIPs"] = [ext_ip]
        if ipv6:
          service.spec["ipFamilies"] = ["IPv6"]

        api_response = self.cluster.create_namespaced_service(
            body=service,
            namespace=ns,
        )
        logger.debug("Additional Service created. status='%s'" % str(api_response.status))

    def wait_until_exists(self, name, resource_type, ns="default"):
        retry_until_success(kubectl, function_args=["get %s %s -n %s" %
                                                    (resource_type, name, ns)])

    def delete_and_confirm(self, name, resource_type, ns="default"):
        try:
            kubectl("delete %s %s -n %s" % (resource_type, name, ns))
        except subprocess.CalledProcessError:
            pass

        def is_it_gone_yet(res_name, res_type):
            try:
                kubectl("get %s %s -n %s" % (res_type, res_name, ns),
                        logerr=False)
                raise self.StillThere
            except subprocess.CalledProcessError:
                # Success
                pass

        retry_until_success(is_it_gone_yet, retries=10, wait_time=10, function_args=[name, resource_type])

    class StillThere(Exception):
        pass

    def get_routes(self):
        return run("docker exec kube-node-extra ip r")

    def update_ds_env(self, ds, ns, env_vars):
        config.load_kube_config(os.environ.get('KUBECONFIG'))
        api = client.AppsV1Api(client.ApiClient())
        node_ds = api.read_namespaced_daemon_set(ds, ns, exact=True, export=False)
        for container in node_ds.spec.template.spec.containers:
            if container.name == ds:
                for k, v in env_vars.items():
                    logger.info("Set %s=%s", k, v)
                    env_present = False
                    for env in container.env:
                        if env.name == k:
                            env_present = True
                    if not env_present:
                        v1_ev = client.V1EnvVar(name=k, value=v, value_from=None)
                        container.env.append(v1_ev)
        api.replace_namespaced_daemon_set(ds, ns, node_ds)

        # Wait until the DaemonSet reports that all nodes have been updated.
        while True:
            time.sleep(10)
            node_ds = api.read_namespaced_daemon_set_status("calico-node", "calico-system")
            logger.info("%d/%d nodes updated",
                      node_ds.status.updated_number_scheduled,
                      node_ds.status.desired_number_scheduled)
            if node_ds.status.updated_number_scheduled == node_ds.status.desired_number_scheduled:
                break

    def scale_deployment(self, deployment, ns, replicas):
        return kubectl("scale deployment %s -n %s --replicas %s" %
                       (deployment, ns, replicas)).strip()

    def get_calico_node_pod(self, nodeName):
        """Get the calico-node pod name for a given kind node"""
        def fn():
            calicoPod = kubectl("-n calico-system get pods -o wide | grep calico-node | grep '%s '| cut -d' ' -f1" % nodeName)
            if calicoPod is None:
                raise Exception('calicoPod is None')
            return calicoPod.strip()
        calicoPod = retry_until_success(fn)
        return calicoPod

class Container(object):

    def __init__(self, image, args, flags=""):
        self.id = run("docker run --rm -d --net=kind %s %s %s" % (
            flags,
            image,
            args)).strip().split("\n")[-1].strip()
        self._ip = None

    def kill(self):
        run("docker rm -f %s" % self.id)

    def inspect(self, template):
        return run("docker inspect -f '%s' %s" % (template, self.id))

    def running(self):
        return self.inspect("{{.State.Running}}").strip()

    def assert_running(self):
        assert self.running() == "true"

    def wait_running(self):
        retry_until_success(self.assert_running)

    @property
    def ip(self):
        if not self._ip:
            self._ip = self.inspect(
                "{{range .NetworkSettings.Networks}}{{.IPAddress}}{{end}}"
            ).strip()
        return self._ip

    def logs(self):
        return run("docker logs %s 2>&1" % self.id)

    def execute(self, cmd):
        return run("docker exec %s %s" % (self.id, cmd))


class Pod(object):

    def __init__(self, ns, name, node=None, image=None, labels=None, annotations=None, yaml=None, cmd=None):
        if yaml:
            # Caller has provided the complete pod YAML.
            kubectl("""apply -f - <<'EOF'
%s
EOF
""" % yaml)
        else:
            # Build YAML with specified namespace, name and image.
            pod = {
                "apiVersion": "v1",
                "kind": "Pod",
                "metadata": {
                    "name": name,
                    "namespace": ns,
                },
                "spec": {
                    "containers": [
                        {
                            "name": name,
                            "image": image,
                        },
                    ],
                    "terminationGracePeriodSeconds": 0,
                },
            }
            if node:
                pod["spec"]["nodeName"] = node
            if annotations:
                pod["metadata"]["annotations"] = annotations
            if labels:
                pod["metadata"]["labels"] = labels
            if cmd:
                pod["spec"]["containers"][0]["command"] = cmd
            kubectl("""apply -f - <<'EOF'
%s
EOF
""" % json.dumps(pod))

        self.name = name
        self.ns = ns
        self._ip = None
        self._ipv6 = None
        self._hostip = None
        self._nodename = None

    def delete(self):
        kubectl("delete pod/%s -n %s" % (self.name, self.ns))

    def wait_ready(self):
        kubectl("wait --for=condition=ready pod/%s -n %s --timeout=300s" % (self.name, self.ns))

    def wait_not_ready(self):
        kubectl("wait --for=condition=Ready=false pod/%s -n %s --timeout=300s" % (self.name, self.ns))

    @property
    def ip(self):
        start_time = time.time()
        while not self._ip:
            assert time.time() - start_time < 30, "Pod failed to get IP address within 30s"
            ip = run("kubectl get po %s -n %s -o json | jq '.status.podIP'" % (self.name, self.ns)).strip().strip('"')
            if ip != "null":
                self._ip = ip
                break
            time.sleep(0.1)
        return self._ip

    @property
    def ipv6(self):
        start_time = time.time()
        while not self._ipv6:
            assert time.time() - start_time < 30, "Pod failed to get IP address within 30s"
            ips = run("kubectl get po %s -n %s -o=jsonpath='{.status.podIPs[*].ip}'" % (self.name, self.ns)).strip().split(" ")
            for ip in ips:
                if ":" in ip:
                    self._ipv6 = ip
                    break
            time.sleep(0.1)
        return self._ipv6

    @property
    def hostip(self):
        if not self._hostip:
            self._hostip = run("kubectl get po %s -n %s -o json | jq '.status.hostIP'" %
                           (self.name, self.ns)).strip().strip('"')
        return self._hostip

    @property
    def nodename(self):
        if not self._nodename:
            # spec.nodeName will be populated for a running pod regardless of being specified or not on pod creation.
            self._nodename = run("kubectl get po %s -n %s -o json | jq '.spec.nodeName'" %
                               (self.name, self.ns)).strip().strip('"')
        return self._nodename

    @property
    def annotations(self):
        return json.loads(run("kubectl get po %s -n %s -o json | jq '.metadata.annotations'" %
                           (self.name, self.ns)).strip().strip('"'))

    def execute(self, cmd, timeout=0):
        return kubectl("exec %s -n %s -- %s" % (self.name, self.ns, cmd), timeout=timeout)

class TestBaseV6(TestBase):

    def get_routes(self):
        return run("docker exec kube-node-extra ip -6 r")



import logging
import SocketServer
import time

logging.basicConfig(level=logging.DEBUG,
                    format='%(name)s: %(message)s',
                    )

logger = logging.getLogger(__name__)


class EchoRequestHandlerTCP(SocketServer.BaseRequestHandler):
    def handle(self):
        logger.debug('handle')
        # Echo the back to the client
        data = self.request.recv(1024)
        logger.debug('received (tcp) from %s: "%s"',
                     self.client_address, data)
        self.request.send(data)
        return


class EchoRequestHandlerUDP(SocketServer.BaseRequestHandler):
    def handle(self):
        logger.debug('handle')

        # Echo the back to the client
        data = self.request[0]
        socket = self.request[1]
        logger.debug('received (udp) from %s: "%s"',
                     self.client_address, data)
        socket.sendto(data, self.client_address)
        return


class EchoServerTCP(SocketServer.TCPServer):
    def serve_forever(self):
        logger.info('waiting for tcp request')
        while True:
            self.handle_request()
        return


class EchoServerUDP(SocketServer.UDPServer):
    def serve_forever(self):
        logger.info('waiting for udp request')
        while True:
            self.handle_request()
        return


if __name__ == '__main__':
    import socket
    import threading

    def check_socket(sock):
        # Send the data
        message = 'Hello world'
        logger.debug('sending data: "%s"', message)
        len_sent = sock.send(message)

        # Receive a response
        logger.debug('waiting for response')
        response = sock.recv(len_sent)
        logger.debug('response from server: "%s"', response)

    tcp_addr = "0.0.0.0"
    tcp_port = 80
    udp_addr = "0.0.0.0"
    udp_port = 69

    tcp_server = EchoServerTCP((tcp_addr, tcp_port), EchoRequestHandlerTCP)
    udp_server = EchoServerUDP((udp_addr, udp_port), EchoRequestHandlerUDP)

    try:
        t1 = threading.Thread(target=tcp_server.serve_forever)
        t1.setDaemon(True)  # don't hang on exit
        t1.start()
        t2 = threading.Thread(target=udp_server.serve_forever)
        t2.setDaemon(True)  # don't hang on exit
        t2.start()

        logger.info('TCP Server on %s:%s', tcp_addr, tcp_port)
        logger.info('UDP Server on %s:%s', udp_addr, udp_port)

        logger.debug('checking tcp server')
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        logger.debug('connecting to server')
        s.connect((tcp_addr, tcp_port))
        check_socket(s)
        s.close()

        logger.debug('checking udp server')
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        logger.debug('connecting to server')
        s.connect((udp_addr, udp_port))
        check_socket(s)
        s.close()
        while True:
            time.sleep(10)
    finally:
        # Clean up
        logger.debug('done')
        tcp_server.socket.close()
        udp_server.socket.close()
        logger.debug('closed sockets')



# Copyright (c) 2025 Tigera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import random
import subprocess
import time
import enum

from tests.k8st.test_base import TestBase, Pod
from tests.k8st.utils.utils import start_external_node_with_bgp, \
    run, calicoctl, kubectl, node_info, retry_until_success, calico_node_pod_name

_log = logging.getLogger(__name__)

class TopologyMode(enum.Enum):
    RR = "rr"
    MESH = "mesh"

bird_conf_tor_mesh = """
# Template for all BGP clients
template bgp bgp_template {
  debug { states };
  description "Connection to BGP peer";
  local as 63000;
  multihop;
  gateway recursive; # This should be the default, but just in case.
  import all;        # Import all routes, since we don't know what the upstream
                     # topology is and therefore have to trust the ToR/RR.
  export all;
  source address ip@local;  # The local address we use for the TCP connection
  add paths on;
  graceful restart;  # See comment in kernel section about graceful restart.
  connect delay time 2;
  connect retry time 5;
  error wait time 5,30;
}

# ------------- Node-to-node mesh -------------
protocol bgp Mesh_with_master_node from bgp_template {
  neighbor %s as 64512;
  passive on;
}

protocol bgp Mesh_with_node_1 from bgp_template {
  neighbor %s as 64512;
  passive on;
}

protocol bgp Mesh_with_node_2 from bgp_template {
  neighbor %s as 64512;
  passive on;
}

protocol bgp Mesh_with_node_3 from bgp_template {
  neighbor %s as 64512;
  passive on;
}
"""

bird_conf_tor_rr = """
# Template for all BGP clients
template bgp bgp_template {
  debug { states };
  description "Connection to BGP peer";
  local as 63000;
  multihop;
  gateway recursive; # This should be the default, but just in case.
  import all;        # Import all routes, since we don't know what the upstream
                     # topology is and therefore have to trust the ToR/RR.
  export all;
  source address ip@local;  # The local address we use for the TCP connection
  add paths on;
  graceful restart;  # See comment in kernel section about graceful restart.
  connect delay time 2;
  connect retry time 5;
  error wait time 5,30;
}

# ------------- RR -------------
protocol bgp RR_with_master_node from bgp_template {
  neighbor %s as 64512;
  passive on;
}
"""


class _TestLocalBGPPeer(TestBase):
    def set_topology(self, value):
        self.topology = value

    def setUp(self):
        super(_TestLocalBGPPeer, self).setUp()

        if self.topology == TopologyMode.MESH:
          _log.info("Topology MESH")
        elif self.topology == TopologyMode.RR:
          _log.info("Topology RR")
        else:
          _log.exception("Topology unknown")

        # Create bgp test namespace
        self.ns = "bgp-test-" + hex(random.randint(0, 0xffffffff))
        self.create_namespace(self.ns)
        self.add_cleanup(lambda: self.delete_and_confirm(self.ns, "ns"))

        self.nodes, self.ips, _ = node_info()

        # Clean up any existing external node just in case it was left over from
        # a previous run.
        self.delete_extra_node()

        # Create the ToR node.
        self.external_node_ip = start_external_node_with_bgp(
            "kind-node-tor",
            bird_peer_config=self.get_bird_conf_tor(),
        )
        self.add_cleanup(self.delete_extra_node)

        # Create the BGP filter to export to the ToR.
        kubectl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPFilter
metadata:
  name: export-child-cluster-cidr
spec:
  exportV4:
  - action: Accept
    matchOperator: In
    cidr: 10.123.0.0/16
    source: RemotePeers
  exportV6:
  - action: Accept
    matchOperator: In
    cidr: ca11:c0::/32
    source: RemotePeers
""")
        self.add_cleanup(lambda: calicoctl("delete bgpfilter export-child-cluster-cidr", allow_fail=True))

        if self.topology == TopologyMode.MESH:
          # Establish BGPPeer from cluster nodes to tor
          kubectl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPPeer
metadata:
  name: node-tor-peer
spec:
  peerIP: %s
  asNumber: 63000
  filters:
  - export-child-cluster-cidr
EOF
""" % self.external_node_ip)
          self.add_cleanup(lambda: calicoctl("delete bgppeer node-tor-peer", allow_fail=True))

        if self.topology == TopologyMode.RR:
          # Configure kind-control-plane to be a route reflector (rr)
          kubectl("annotate node kind-control-plane projectcalico.org/RouteReflectorClusterID=244.0.0.1")
          self.add_cleanup(lambda: kubectl("annotate node kind-control-plane projectcalico.org/RouteReflectorClusterID-"))

          # Configure other nodes to peer with rr
          kubectl("""apply -f - << EOF
kind: BGPPeer
apiVersion: projectcalico.org/v3
metadata:
  name: peer-with-rr
spec:
  nodeSelector: all()
  peerSelector: kubernetes.io/hostname == 'kind-control-plane'
  nextHopMode: Self
  reversePeering: Manual
  filters:
  - export-child-cluster-cidr
""")
          self.add_cleanup(lambda: calicoctl("delete bgppeer peer-with-rr", allow_fail=True))

          # Configure rr to peer with other nodes
          kubectl("""apply -f - << EOF
kind: BGPPeer
apiVersion: projectcalico.org/v3
metadata:
  name: peer-from-rr
spec:
  peerSelector: all()
  nodeSelector: kubernetes.io/hostname == 'kind-control-plane'
  reversePeering: Manual
  filters:
  - export-child-cluster-cidr
""")
          self.add_cleanup(lambda: calicoctl("delete bgppeer peer-from-rr", allow_fail=True))

          # Establish BGPPeer from rr to tor
          kubectl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPPeer
metadata:
  name: rr-tor-peer
spec:
  nodeSelector: kubernetes.io/hostname == 'kind-control-plane'
  peerIP: %s
  asNumber: 63000
  nextHopMode: Keep
  filters:
  - export-child-cluster-cidr
EOF
""" % self.external_node_ip)
          self.add_cleanup(lambda: calicoctl("delete bgppeer rr-tor-peer", allow_fail=True))


        # Create three pods. Name format : color-pod-host-sequence number
        # kind-worker node starts from nodes[1]
        self.red_pod_0_0 = self.create_workload_pod(self.nodes[1], "red-pod-0-0", self.ns, "red")
        self.red_pod_1_0 = self.create_workload_pod(self.nodes[2], "red-pod-1-0", self.ns, "red")
        self.blue_pod_0_0 = self.create_workload_pod(self.nodes[1], "blue-pod-0-0", self.ns, "blue")
        self.blue_pod_1_0 = self.create_workload_pod(self.nodes[2], "blue-pod-1-0", self.ns, "blue")

        for (p, block_v4, block_v6) in [
            (self.red_pod_0_0,"10.123.0.0/26","ca11:c0::/96"),
            (self.red_pod_1_0,"10.123.1.0/26","ca11:c0:1::/96"),
            (self.blue_pod_0_0,"10.123.2.0/26","ca11:c0:2::/96"),
            (self.blue_pod_1_0,"10.123.3.0/26","ca11:c0:3::/96"),
        ]:
            p.wait_ready()
            self.setup_workload_pod_v4(p, block_v4)
            self.setup_workload_pod_v6(p, block_v6)

        # Define localWorkloadPeeringIP and turn off nodeToNodeMesh
        kubectl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPConfiguration
metadata:
 name: default
spec:
  localWorkloadPeeringIPV4: 169.254.0.179
  localWorkloadPeeringIPV6: fd12:3456:789a::1
  nodeToNodeMeshEnabled: false
  asNumber: 64512
""")
        self.add_cleanup(lambda: kubectl("delete bgpconfiguration default"))

        # Create the BGP filters for the local peerings.  We want to aviod
        # exporting routes to the local peers since they already have a
        # default route.
        kubectl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPFilter
metadata:
  name: accept-pod-cidr-child-cluster
spec:
  importV4:
    - action: Accept
      matchOperator: In
      cidr: 10.123.0.0/16
  importV6:
    - action: Accept
      matchOperator: In
      cidr: ca11:c0::/32
---
apiVersion: projectcalico.org/v3
kind: BGPFilter
metadata:
  name: no-export-from-parent-cluster
spec:
  exportV4:
    - action: Reject
  exportV6:
    - action: Reject
""")

        # Create the local peerings.
        kubectl("""apply -f - << EOF
apiVersion: projectcalico.org/v3
kind: BGPPeer
metadata:
 name: global-peer
spec:
 localWorkloadSelector: color == 'red'
 asNumber: 65401
 filters:
 - no-export-from-parent-cluster
 - accept-pod-cidr-child-cluster
---
apiVersion: projectcalico.org/v3
kind: BGPPeer
metadata:
 name: node-peer
spec:
 nodeSelector: kubernetes.io/hostname == 'kind-worker2'
 localWorkloadSelector: color == 'blue'
 asNumber: 65401
 filters:
 - no-export-from-parent-cluster
 - accept-pod-cidr-child-cluster
EOF
""")
        self.add_cleanup(lambda: calicoctl("delete bgppeer global-peer", allow_fail=True))
        self.add_cleanup(lambda: calicoctl("delete bgppeer node-peer", allow_fail=True))

    @staticmethod
    def delete_extra_node():
        try:
            # Delete the extra node.
            run("docker rm -f kind-node-tor")
        except subprocess.CalledProcessError:
            pass

    def create_workload_pod(self, host, name, ns="default", color="red"):
        """
        Create pod as workload peer.
        """

        pod = Pod(ns, name, image=None, yaml="""
apiVersion: v1
kind: Pod
metadata:
  name: %s
  namespace: %s
  labels:
    color: %s
spec:
  containers:
  - name: bird
    image: calico/bird:v0.3.3-211-g9111ec3c
    securityContext:
      privileged: true
  nodeName: %s
  terminationGracePeriodSeconds: 0
""" % (name, ns, color, host))
        self.add_cleanup(pod.delete)
        return pod

    def setup_workload_pod_v4(self, pod, ipam_block):
        # Copy bird
        bird_peer_config = self.get_bird_config_workload("65401", pod.ip,pod.ip, ipam_block, "169.254.0.179", "64512")
        with open('peers.conf', 'w') as peerconfig:
            peerconfig.write(bird_peer_config)

        run("kubectl cp peers.conf %s/%s:/etc/bird.conf" % (pod.ns, pod.name))
        run("kubectl exec -t %s -n %s -- sh -c 'birdcl configure'" % (pod.name, pod.ns))

        run("rm peers.conf")

        # run("kubectl exec -t %s -n %s -- sh -c 'ip route add 10.244.0.64/24 via 192.168.0.8'" % (pod.name, pod.ns))

    def setup_workload_pod_v6(self, pod, ipam_block):
        # Copy bird
        bird_peer_config = self.get_bird_config_workload("65401", pod.ip,pod.ipv6, ipam_block, "fd12:3456:789a::1", "64512")
        with open('peers.conf', 'w') as peerconfig:
            peerconfig.write(bird_peer_config)

        run("kubectl cp peers.conf %s/%s:/etc/bird6.conf" % (pod.ns, pod.name))
        run("kubectl exec -t %s -n %s -- sh -c 'birdcl6 configure'" % (pod.name, pod.ns))

        run("rm peers.conf")

    def get_bird_conf_tor(self):
        if self.topology == TopologyMode.MESH:
          return bird_conf_tor_mesh % (self.ips[0], self.ips[1], self.ips[2], self.ips[3])
        if self.topology == TopologyMode.RR:
          return bird_conf_tor_rr % self.ips[0]

    def get_bird_config_workload(self, as_number_child, router_id, child_ip, child_block, local_workload_peer_ip,
                                 as_number_parent):
        return """
router id %s;

function calico_cidr_filter() {
  if ( net ~ %s ) then {
    accept;
  }
}

# Configure synchronization between routing tables and kernel.
protocol kernel {
  learn;             # Learn all alien routes from the kernel
  persist;           # Don't remove routes on bird shutdown
  scan time 2;       # Scan kernel routing table every 2 seconds
  import all;
  export filter {
    calico_cidr_filter();
    reject;
  };
  graceful restart;  # Turn on graceful restart to reduce potential flaps in
                     # routes when reloading BIRD configuration.  With a full
                     # automatic mesh, there is no way to prevent BGP from
                     # flapping since multiple nodes update their BGP
                     # configuration at the same time, GR is not guaranteed to
                     # work correctly in this scenario.
  merge paths on;    # Allow export multipath routes (ECMP)
}

# Watch interface up/down events.
protocol device {
  debug { states };
  scan time 2;    # Scan interfaces every 2 seconds
}

# A static route to export to the parent cluster.
protocol static {
    route %s via %s;
}

# Template for all BGP clients
template bgp bgp_template {
  debug { states };
  description "Connection to BGP peer";
  local as %s;
  gateway recursive; # This should be the default, but just in case.
  add paths on;
  graceful restart;  # See comment in kernel section about graceful restart.
  connect delay time 2;
  connect retry time 5;
  error wait time 5,30;
  import filter {
    calico_cidr_filter();
    reject;
  };
  export filter {
    calico_cidr_filter();
    reject;
  };
  ttl security off;
  multihop;
}

protocol bgp from_workload_to_local_host from bgp_template {
  neighbor %s as %s;
}
""" % (router_id, child_block,  child_block, child_ip, as_number_child, local_workload_peer_ip, as_number_parent)

    # Given a pod, check if should have BGP connections with the host.
    def assert_bgp_established(self, pod):
        output = run("kubectl exec -t %s -n %s -- birdcl show protocols" % (pod.name, pod.ns))
        regexp = "from_workload_to_local_host.*Established"
        self.assertRegex(output, regexp, "IPv4 BGP connection not established, pod " + pod.name)
        output = run("kubectl exec -t %s -n %s -- birdcl6 show protocols" % (pod.name, pod.ns))
        self.assertRegex(output, regexp, "IPv6 BGP connection not established, pod " + pod.name)

    def assert_bgp_not_established(self, pod):
        output = run("kubectl exec -t %s -n %s -- birdcl show protocols" % (pod.name, pod.ns))
        regexp = "from_workload_to_local_host.*Established"
        self.assertNotRegex(output, regexp, "IPv4 BGP connection unexpectedly established, pod " + pod.name)
        output = run("kubectl exec -t %s -n %s -- birdcl6 show protocols" % (pod.name, pod.ns))
        self.assertNotRegex(output, regexp, "IPv6 BGP connection unexpectedly established, pod " + pod.name)

    def _test_local_bgp_peers(self):
        """
        Runs the tests for local bgp peers
        """
        stop_for_debug()

        # Assert bgp sessions has been established to the following local workloads.
        # red pods on kind-worker and kind-worker2. blue pod on kind-worker2.
        retry_until_success(self.assert_bgp_established, function_args=[self.red_pod_0_0], retries=60)
        retry_until_success(self.assert_bgp_established, function_args=[self.red_pod_1_0], retries=60)
        self.assert_bgp_not_established(self.blue_pod_0_0)
        retry_until_success(self.assert_bgp_established, function_args=[self.blue_pod_1_0], retries=60)

        # Check the export filter is applied.  Child nodes shouldn't see routes
        # from the parent.
        output = run("kubectl exec -t %s -n %s -- birdcl show route" % (self.red_pod_0_0.name, self.red_pod_0_0.ns))
        self.assertRegex(output, r"0\.0\.0\.0.*via 169.254.1.1")
        self.assertNotRegex(output, r"192\.168\.\d+\.\d+/26", "Unexpected route to parent cluster")

        # Check that the import filter accepts child routes.
        calico_node_w1 = calico_node_pod_name(self.red_pod_0_0.nodename)
        # Expect a single route like this for worker 1.
        # 10.123.0.0/26      via 192.168.162.157 on calicef4c701383 [Local_Workload_192_168_162_157 15:11:46] * (100/0) [AS65401i]
        output = run("kubectl exec -t %s -n calico-system -- birdcl show route" % calico_node_w1)
        self.assertRegex(output, r"10\.123\.0\.0/26.*via .* on cali.*Local_Workload_.*AS65401")
        output = run("kubectl exec -t %s -n calico-system -- birdcl6 show route" % calico_node_w1)
        self.assertRegex(output, r"ca11:c0::/96.*via .* on cali.*Local_Workload_.*AS65401")

        # Worker 2 should have 2 routes of each version, one for each child.
        calico_node_w2 = calico_node_pod_name(self.red_pod_1_0.nodename)
        output = run("kubectl exec -t %s -n calico-system -- birdcl show route" % calico_node_w2)
        self.assertRegex(output, r"10\.123\.1\.0/26.*via .* on cali.*Local_Workload_.*AS65401")
        self.assertRegex(output, r"10\.123\.3\.0/26.*via .* on cali.*Local_Workload_.*AS65401")
        output = run("kubectl exec -t %s -n calico-system -- birdcl6 show route" % calico_node_w2)
        self.assertRegex(output, r"ca11:c0:1::/96.*via .* on cali.*Local_Workload_.*AS65401")
        self.assertRegex(output, r"ca11:c0:3::/96.*via .* on cali.*Local_Workload_.*AS65401")

        if self.topology == TopologyMode.MESH:
          # Check that the ToR hears about all the routes.
          # 10.123.0.0/26      via 172.18.0.3 on eth0 [Mesh_with_node_1 17:19:12] * (100/0) [AS65401i]
          # 10.123.1.0/26      via 172.18.0.2 on eth0 [Mesh_with_node_2 17:19:11] * (100/0) [AS65401i]
          # 10.123.3.0/26      via 172.18.0.2 on eth0 [Mesh_with_node_2 17:19:09] * (100/0) [AS65401i]
          output = run("docker exec kind-node-tor birdcl show route")
          self.assertRegex(output, r"10\.123\.0\.0/26.*via %s on .*Mesh_with_node_1.*AS65401" % (self.ips[1],))
          self.assertRegex(output, r"10\.123\.1\.0/26.*via %s on .*Mesh_with_node_2.*AS65401" % (self.ips[2],))
          self.assertRegex(output, r"10\.123\.3\.0/26.*via %s on .*Mesh_with_node_2.*AS65401" % (self.ips[2],))

        if self.topology == TopologyMode.RR:
          # Check that kind-worker3 hears about all the routes from RR(master node) with original next hop.
          # 10.123.3.0/26      via 172.18.0.5 on eth0 [Node_172_18_0_3 09:46:12 from 172.18.0.3] * (100/0) [AS65401i]
          # 10.123.0.0/26      via 172.18.0.2 on eth0 [Node_172_18_0_3 09:46:10 from 172.18.0.3] * (100/0) [AS65401i]
          # 10.123.1.0/26      via 172.18.0.5 on eth0 [Node_172_18_0_3 09:46:12 from 172.18.0.3] * (100/0) [AS65401i]
          calico_node_w3 = calico_node_pod_name(self.nodes[3])
          output = run("kubectl exec -t %s -n calico-system -- birdcl show route" % calico_node_w3)
          self.assertRegex(output, r"10\.123\.0\.0/26.*via %s on .*Node_172_18_0_.*AS65401" % (self.ips[1],))
          self.assertRegex(output, r"10\.123\.1\.0/26.*via %s on .*Node_172_18_0_.*AS65401" % (self.ips[2],))
          self.assertRegex(output, r"10\.123\.3\.0/26.*via %s on .*Node_172_18_0_.*AS65401" % (self.ips[2],))

          # Check that the ToR hears about all the routes from RR(master node).
          # Note that `nextHopMode: Keep` is specified for `rr-tor-peer`, ToR sees routes with original next hop.
          # 10.123.3.0/26      via 172.18.0.5 on eth0 [RR_with_master_node 09:46:12 from 172.18.0.3] * (100/0) [AS65401i]
          # 10.123.0.0/26      via 172.18.0.2 on eth0 [RR_with_master_node 09:46:10 from 172.18.0.3] * (100/0) [AS65401i]
          # 10.123.1.0/26      via 172.18.0.5 on eth0 [RR_with_master_node 09:46:12 from 172.18.0.3] * (100/0) [AS65401i]
          output = run("docker exec kind-node-tor birdcl show route")
          self.assertRegex(output, r"10\.123\.0\.0/26.*via %s on .*RR_with_master_node.*AS65401" % (self.ips[1],))
          self.assertRegex(output, r"10\.123\.1\.0/26.*via %s on .*RR_with_master_node.*AS65401" % (self.ips[2],))
          self.assertRegex(output, r"10\.123\.3\.0/26.*via %s on .*RR_with_master_node.*AS65401" % (self.ips[2],))

        # Check connectivity from ToR to workload.
        self.red_pod_0_0.execute("ip addr add 10.123.0.1 dev lo")

        output = run("docker exec kind-node-tor ping -c3 10.123.0.1")
        self.assertRegex(output, "3 packets transmitted, 3 packets received")

class TestLocalBGPPeerRR(_TestLocalBGPPeer):

    # In the tests of this class we have BGP peers between the
    # cluster nodes (kind-control-plane, kind-worker, kind-worker2, kind-worker3, kind-control-plane acting as a RR) with ASNumber 64512
    # and the external node (kind-node-tor ASNumber 63000). We test BGP connections between local
    # workload peers (ASNumber 65401) and the cluster nodes.
    #
    # - The full mesh between the cluster nodes is turned off by
    #   nodeToNodeMeshEnabled: false.
    #
    # - Two pods (red & blue) are deployed on kind-worker node and two pods (red & blue)
    #   are deployed on kind-worker2 node.
    #
    # - A global peer is created to select red pods as local bgp peers.
    #   A node specific peer is created to select the blue pod on kind-worker2 node.

    def setUp(self):
        self.set_topology(TopologyMode.RR)
        super(TestLocalBGPPeerRR, self).setUp()

    def test_local_bgp_peers(self):
        self._test_local_bgp_peers()


class TestLocalBGPPeerMesh(_TestLocalBGPPeer):

    # In the tests of this class we have BGP peers between the
    # cluster nodes (kind-control-plane, kind-worker, kind-worker2, kind-worker3) with ASNumber 64512
    # and the external node (kind-node-tor ASNumber 63000). We test BGP connections between local
    # workload peers (ASNumber 65401) and the cluster nodes.
    #
    # - The full mesh between the cluster nodes is turned off by
    #   nodeToNodeMeshEnabled: false.
    #
    # - Two pods (red & blue) are deployed on kind-worker node and two pods (red & blue)
    #   are deployed on kind-worker2 node.
    #
    # - A global peer is created to select red pods as local bgp peers.
    #   A node specific peer is created to select the blue pod on kind-worker2 node.

    def setUp(self):
        self.set_topology(TopologyMode.MESH)
        super(TestLocalBGPPeerMesh, self).setUp()

    def test_local_bgp_peers(self):
        self._test_local_bgp_peers()


def stop_for_debug():
    # Touch debug file under projectcalico/calico/node to stop the process
    debug_file = "/code/stop-for-debug"
    last_log = None
    while os.path.exists(debug_file):
        if last_log is None or time.time() - last_log > 60:
            _log.info("stop-for-debug file exists. Sleeping...")
            last_log = time.time()
        time.sleep(1)

    _log.info("Debug file does not exist. Continuing...")



# -*- coding: utf-8 -*-
# Copyright (c) 2018 Tigera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test compaction code.
"""

import logging
import os
import unittest

from etcd3gw.exceptions import Etcd3Exception

import mock

import networking_calico.plugins.ml2.drivers.calico.test.lib as lib
from networking_calico import etcdv3
from networking_calico.plugins.ml2.drivers.calico import mech_calico


LOG = logging.getLogger(__name__)


class MockLease(object):

    def __init__(self, id, client):
        self.id = id

    def ttl(self):
        if self.id == 1:
            # A reasonable TTL.
            return 30
        else:
            # An unreasonably large TTL.
            return 30000000


class TestCompaction(unittest.TestCase):

    def setUp(self):
        super(TestCompaction, self).setUp()

        # Hook the etcd3gw client.
        etcdv3._client = self.client = mock.Mock()
        # Writes succeed by default.
        self.client.put.return_value = True
        # Insinuate our mock lease class.
        etcdv3.Lease = MockLease

        # Hook relevant logging.
        import logging

        for module in [
            mech_calico,
            etcdv3,
        ]:
            module.LOG = logging.getLogger("\t%-15s\t" % module.__name__.split(".")[-1])

        self.pid = str(os.getpid())

        # Provide default config.
        lib.m_oslo_config.cfg.CONF.calico.etcd_compaction_period_mins = 60
        lib.m_oslo_config.cfg.CONF.calico.etcd_compaction_min_revisions = 1000

    def tearDown(self):
        etcdv3._client = None
        super(TestCompaction, self).tearDown()

    def test_iter_1(self):
        LOG.info("1. no compaction keys present")
        self.client.get.side_effect = iter(
            [
                # Read trigger: not found.
                [],
                # Read last: not found.
                [],
            ]
        )
        mech_calico.check_request_etcd_compaction()
        self.assertEqual(
            [
                mock.call("/calico/compaction/v1/last", "0", lease=None),
                mock.call("/calico/compaction/v1/trigger", self.pid, lease=mock.ANY),
            ],
            self.client.put.mock_calls,
        )
        self.client.post.assert_not_called()

    def test_iter_2(self):
        LOG.info("2. compacted@0 checked@10 now=100 => no compaction")
        self.client.get.side_effect = iter(
            [
                # Read trigger: not found.
                [],
                # Read last.
                [("0".encode(), {"mod_revision": "10"})],
            ]
        )
        self.client.status.return_value = {
            "header": {
                "cluster_id": "12345",
                "revision": "100",
            }
        }
        mech_calico.check_request_etcd_compaction()
        self.assertEqual(
            [
                mock.call("/calico/compaction/v1/last", "0", lease=None),
                mock.call("/calico/compaction/v1/trigger", self.pid, lease=mock.ANY),
            ],
            self.client.put.mock_calls,
        )
        self.client.post.assert_not_called()

    def test_iter_3(self):
        LOG.info("3. compacted@0 checked@100 now=300 => no compaction")
        self.client.get.side_effect = iter(
            [
                # Read trigger: not found.
                [],
                # Read last.
                [("0".encode(), {"mod_revision": "100"})],
            ]
        )
        self.client.status.return_value = {
            "header": {
                "cluster_id": "12345",
                "revision": "300",
            }
        }
        mech_calico.check_request_etcd_compaction()
        self.assertEqual(
            [
                mock.call("/calico/compaction/v1/last", "0", lease=None),
                mock.call("/calico/compaction/v1/trigger", self.pid, lease=mock.ANY),
            ],
            self.client.put.mock_calls,
        )
        self.client.post.assert_not_called()

    def test_iter_4(self):
        LOG.info("4. compacted@0 checked@300 now=1100 => compact@100")
        self.client.get.side_effect = iter(
            [
                # Read trigger: not found.
                [],
                # Read last.
                [("0".encode(), {"mod_revision": "300"})],
            ]
        )
        self.client.status.return_value = {
            "header": {
                "cluster_id": "12345",
                "revision": "1100",
            }
        }
        mech_calico.check_request_etcd_compaction()
        self.assertEqual(
            [
                mock.call("/calico/compaction/v1/last", "100", lease=None),
                mock.call("/calico/compaction/v1/trigger", self.pid, lease=mock.ANY),
            ],
            self.client.put.mock_calls,
        )
        self.client.post.assert_called()

    def test_iter_5(self):
        LOG.info("5. compacted@100 checked@1100 now=1200 => compact@200")
        self.client.get.side_effect = iter(
            [
                # Read trigger: not found.
                [],
                # Read last.
                [("100".encode(), {"mod_revision": "1100"})],
            ]
        )
        self.client.status.return_value = {
            "header": {
                "cluster_id": "12345",
                "revision": "1200",
            }
        }
        mech_calico.check_request_etcd_compaction()
        self.assertEqual(
            [
                mock.call("/calico/compaction/v1/last", "200", lease=None),
                mock.call("/calico/compaction/v1/trigger", self.pid, lease=mock.ANY),
            ],
            self.client.put.mock_calls,
        )
        self.client.post.assert_called()

    def test_trigger_present(self):
        LOG.info("Trigger present; nothing happens")
        self.client.get.side_effect = iter(
            [
                # Read trigger: present with good lease.
                [(self.pid.encode(), {"mod_revision": "2000", "lease": "1"})],
            ]
        )
        mech_calico.check_request_etcd_compaction()
        self.client.status.assert_not_called()
        self.client.put.assert_not_called()
        self.client.post.assert_not_called()

    def test_trigger_present_missing_lease(self):
        LOG.info("Trigger present but with missing lease")
        self.client.get.side_effect = iter(
            [
                # Read trigger: present but lease missing.
                [(self.pid.encode(), {"mod_revision": "2000"})],
            ]
        )
        mech_calico.check_request_etcd_compaction()
        self.assertEqual(
            [
                mock.call("/calico/compaction/v1/last", "0", lease=None),
                mock.call("/calico/compaction/v1/trigger", self.pid, lease=mock.ANY),
            ],
            self.client.put.mock_calls,
        )
        self.client.post.assert_not_called()

    def test_trigger_present_bad_lease(self):
        LOG.info("Trigger present but with bad lease")
        self.client.get.side_effect = iter(
            [
                # Read trigger: present but lease has unreasonably large TTL.
                [(self.pid.encode(), {"mod_revision": "2000", "lease": "2"})],
            ]
        )
        mech_calico.check_request_etcd_compaction()
        self.assertEqual(
            [
                mock.call("/calico/compaction/v1/last", "0", lease=None),
                mock.call("/calico/compaction/v1/trigger", self.pid, lease=mock.ANY),
            ],
            self.client.put.mock_calls,
        )
        self.client.post.assert_not_called()

    def test_bogus_last_compaction_rev(self):
        LOG.info("Bogus last compaction revision > current revision")
        self.client.get.side_effect = iter(
            [
                # Read trigger: not found.
                [],
                # Read last.
                [("3000".encode(), {"mod_revision": "1100"})],
            ]
        )
        self.client.status.return_value = {
            "header": {
                "cluster_id": "12345",
                "revision": "1200",
            }
        }
        mech_calico.check_request_etcd_compaction()
        self.assertEqual(
            [
                mock.call("/calico/compaction/v1/last", "0", lease=None),
                mock.call("/calico/compaction/v1/trigger", self.pid, lease=mock.ANY),
            ],
            self.client.put.mock_calls,
        )
        self.client.post.assert_not_called()

    def test_exception_detail_logging(self):
        LOG.info("Logging detail for etcd3 failure exception")

        # This is like test_iter_5, except we arrange for the
        # etcdv3.request_compaction call to raise an exception, and
        # check the resulting logging.
        e3e = Etcd3Exception("revision has been compacted")
        self.client.post.side_effect = e3e
        self.client.get.side_effect = iter(
            [
                # Read trigger: not found.
                [],
                # Read last.
                [("100".encode(), {"mod_revision": "1100"})],
            ]
        )
        self.client.status.return_value = {
            "header": {
                "cluster_id": "12345",
                "revision": "1200",
            }
        }

        with mock.patch.object(mech_calico.LOG, "info") as mock_li:
            mech_calico.check_request_etcd_compaction()
            self.assertEqual(
                [
                    mock.call("/calico/compaction/v1/last", "1200", lease=None),
                    mock.call(
                        "/calico/compaction/v1/trigger", self.pid, lease=mock.ANY
                    ),
                ],
                self.client.put.mock_calls,
            )
            self.client.post.assert_called()
            mock_li.assert_called_with(
                "Someone else has requested etcd compaction:\n%s",
                "revision has been compacted",
            )



# -*- coding: utf-8 -*-
# Copyright 2014, 2015 Metaswitch Networks
# Copyright (c) 2018 Tigera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test election code.
"""
import logging
import unittest

from etcd3gw import exceptions as e3e

import eventlet

import mock

from networking_calico import etcdv3
from networking_calico.plugins.ml2.drivers.calico import election
from networking_calico.plugins.ml2.drivers.calico.test import stub_etcd


LOG = logging.getLogger(__name__)


def eventlet_sleep(time):
    pass


class TestElection(unittest.TestCase):
    def setUp(self):
        super(TestElection, self).setUp()
        self._real_sleep = eventlet.sleep
        eventlet.sleep = eventlet_sleep
        # Stop eventlet from printing our expected NoMoreResults exception
        # to stdout directly.
        self.print_exc_patch = mock.patch("traceback.print_exception", autospec=True)
        self.print_exc_patch.start()
        # Mock calls to sys.exit.
        self.sys_exit_p = mock.patch("sys.exit")
        self.sys_exit_p.start()

    def tearDown(self):
        self.sys_exit_p.stop()
        self.print_exc_patch.stop()
        eventlet.sleep = self._real_sleep
        etcdv3._client = None
        super(TestElection, self).tearDown()

    def test_invalid(self):
        # Test that not elected using defaults.
        with self.assertRaises(ValueError):
            etcdv3._client = stub_etcd.Client()
            elector = election.Elector("test_basic", "/bloop", interval=-1, ttl=15)
            self.assertFalse(elector.master())
            self._wait_and_stop(etcdv3._client, elector)

        with self.assertRaises(ValueError):
            etcdv3._client = stub_etcd.Client()
            elector = election.Elector("test_basic", "/bloop", interval=10, ttl=5)
            self.assertFalse(elector.master())
            self._wait_and_stop(etcdv3._client, elector)

    def _wait_and_stop(self, client, elector):
        # Wait for the client to tell us that all the results have been
        # processed.
        try:
            eventlet.with_timeout(5, client.no_more_results.wait)
        except eventlet.Timeout:
            elector._greenlet.kill(AssertionError("Didn't reach end of results"))
            elector._greenlet.wait()
            raise
        # This should shut down the Elector.
        eventlet.with_timeout(5, elector.stop)
        # The greenlet should be dead already, but just in case, let our
        # client proceed to raise its exception.
        client.stop.send()
        # Double-check there were no failures.
        self.assertIsNone(client.failure, msg=client.failure)

    def test_basic_election(self):
        # Test that not elected using defaults.
        LOG.debug("test_basic_election")
        etcdv3._client = client = stub_etcd.Client()
        client.add_read_result(key="/bloop", value="value")
        elector = election.Elector("test_basic", "/bloop", interval=5, ttl=15)
        self._wait_and_stop(client, elector)
        self.assertFalse(elector.master())

    def test_become_master_first_time(self):
        # Become the master after one round
        LOG.debug("test_become_master_first_time")
        etcdv3._client = client = stub_etcd.Client()
        client.add_read_exception(etcdv3.KeyNotFound())
        client.add_write_exception(None)
        client.add_write_exception(None)
        client.add_write_exception(None)
        elector = election.Elector(
            "test_basic", "/bloop", old_key="/legacy", interval=5, ttl=15
        )
        self._wait_and_stop(client, elector)
        client.assert_key_written("/legacy")

    def test_fail_to_maintain(self):
        # Become the master after one round
        LOG.debug("test_become_master_first_time")
        etcdv3._client = client = stub_etcd.Client()
        client.add_read_exception(etcdv3.KeyNotFound())
        client.add_write_exception(None)
        client.add_write_exception(e3e.ConnectionFailedError())
        elector = election.Elector("test_basic", "/bloop", interval=5, ttl=15)
        self._wait_and_stop(client, elector)

    def test_become_master_multiple_attempts(self):
        # Become the master after one round
        LOG.debug("test_become_master_multiple_circuits")
        for action in ["delete", "expire", "compareAndDelete", "something"]:
            LOG.info("Testing etcd delete event %s", action)
            etcdv3._client = client = stub_etcd.Client()
            client.add_read_result(key="/bloop", value="value")
            client.add_read_result(key="/bloop", value="value")
            client.add_read_result(key="/bloop", value=None, action=action)
            client.add_write_exception(None)
            client.add_write_exception(None)
            elector = election.Elector("test_basic", "/bloop", interval=5, ttl=15)
            self._wait_and_stop(client, elector)

    def test_become_master_implausible(self):
        # Become the master after key vanishes
        LOG.debug("test_become_master_implausible")
        etcdv3._client = client = stub_etcd.Client()
        client.add_read_result(key="/bloop", value="value")
        client.add_read_result(key="/bloop", value="value")
        client.add_read_exception(etcdv3.KeyNotFound())
        client.add_write_result()
        client.add_write_result()
        elector = election.Elector("test_basic", "/bloop", interval=5, ttl=15)
        self._wait_and_stop(client, elector)

    def test_initial_read_exceptions(self):
        LOG.debug("test_initial_read_exceptions")

        etcdv3._client = client = stub_etcd.Client()
        client.add_read_exception(e3e.Etcd3Exception(detail_text="Unauthorised user"))
        client.add_read_exception(e3e.InternalServerError())
        client.add_read_exception(e3e.ConnectionFailedError())
        client.add_read_exception(e3e.PreconditionFailedError())
        elector = election.Elector("test_basic", "/bloop", interval=5, ttl=15)
        self._wait_and_stop(client, elector)

    def test_exception_detail_logging(self):
        LOG.debug("test_exception_detail_logging")

        with mock.patch.object(election.LOG, "warning") as mock_lw:
            etcdv3._client = client = stub_etcd.Client()
            exc = e3e.Etcd3Exception(detail_text="Unauthorised user")
            client.add_read_exception(exc)
            elector = election.Elector("test_basic", "/bloop", interval=5, ttl=15)
            self._wait_and_stop(client, elector)

            # Check that Etcd3Exception detail was logged.
            mock_lw.assert_called_with(
                "Failed to %s - key %s: %r:\n%s",
                "read current master",
                "/bloop",
                exc,
                "Unauthorised user",
            )

    def test_later_exceptions(self):
        LOG.debug("test_later_read_exceptions")

        etcdv3._client = client = stub_etcd.Client()
        client.add_read_result(key="/bloop", value="value")
        client.add_read_exception(e3e.Etcd3Exception())
        client.add_read_result(key="/bloop", value="value")
        client.add_read_exception(e3e.InternalServerError())
        client.add_read_result(key="/bloop", value="value")
        client.add_read_exception(e3e.ConnectionFailedError())
        client.add_read_result(key="/bloop", value="value")
        client.add_read_exception(e3e.PreconditionFailedError())
        elector = election.Elector("test_basic", "/bloop", interval=5, ttl=15)
        self._wait_and_stop(client, elector)

    def test_master_failure(self):
        LOG.debug("test_master_failure")

        etcdv3._client = client = stub_etcd.Client()
        client.add_read_exception(etcdv3.KeyNotFound())
        # Now become the master but fail
        client.add_write_exception(e3e.ConnectionFailedError())
        client.add_read_result(key="/bloop", value="value")
        client.add_read_result(key="/bloop", value=None, action="delete")
        # Now become the master but fail again
        client.add_write_exception(e3e.InternalServerError())
        # Go back to the beginning again.
        client.add_read_result(key="/bloop", value="value")
        client.add_read_result(key="/bloop", value=None, action="delete")
        client.add_write_exception(None)
        client.add_write_exception(None)
        elector = election.Elector("test_basic", "/bloop", interval=5, ttl=15)
        self._wait_and_stop(client, elector)

        # We are no longer the master, after being told to stop.
        self.assertFalse(elector.master())

    @mock.patch("os.path.exists")
    def test_check_master_process_died(self, m_exists):
        m_exists.return_value = False
        etcdv3._client = client = mock.Mock()
        elector = election.Elector("server-id", "/bloop", interval=5, ttl=15)
        # etcd3 transaction returns False because the key is no longer there.
        client.transaction.return_value = {}
        self.assertRaises(
            election.RestartElection, elector._check_master_process, "server-id:1234"
        )
        self.assertEqual(
            [
                mock.call(
                    {
                        "compare": [
                            {
                                "value": "c2VydmVyLWlkOjEyMzQ=",
                                "result": "EQUAL",
                                "key": "L2Jsb29w",
                                "target": "VALUE",
                            }
                        ],
                        "success": [{"request_delete_range": {"key": "L2Jsb29w"}}],
                        "failure": [],
                    }
                )
            ],
            client.transaction.mock_calls,
        )
        client.failure = None
        self._wait_and_stop(client, elector)

    @mock.patch("os.path.exists")
    def test_check_master_process_other_server(self, m_exists):
        m_exists.return_value = False
        etcdv3._client = client = mock.Mock()
        elector = election.Elector("server-id", "/bloop", interval=5, ttl=15)
        elector._check_master_process("other-server:1234")
        self.assertEqual([], client.delete.mock_calls)
        self.assertEqual([], client.transaction.mock_calls)
        client.failure = None
        self._wait_and_stop(client, elector)

    @mock.patch("os.path.exists")
    def test_check_master_process_still_alive(self, m_exists):
        m_exists.return_value = True
        etcdv3._client = client = mock.Mock()
        elector = election.Elector("server-id", "/bloop", interval=5, ttl=15)
        elector._check_master_process("server-id:1234")
        self.assertEqual([], client.delete.mock_calls)
        self.assertEqual([], client.transaction.mock_calls)
        client.failure = None
        self._wait_and_stop(client, elector)
